"""Unit test for String representations."""

from typing import Any
from unittest.mock import patch

from xknx import XKNX
from xknx.cemi import (
    CEMIFrame,
    CEMILData,
    CEMIMessageCode,
    CEMIMPropInfo,
    CEMIMPropReadResponse,
)
from xknx.devices import (
    BinarySensor,
    Climate,
    ClimateMode,
    Cover,
    DateTimeDevice,
    ExposeSensor,
    Fan,
    Light,
    Notification,
    Scene,
    Sensor,
    Switch,
    Weather,
)
from xknx.dpt import DPTArray, DPTBinary
from xknx.exceptions import (
    ConversionError,
    CouldNotParseAddress,
    CouldNotParseKNXIP,
    CouldNotParseTelegram,
    DeviceIllegalValue,
    IncompleteKNXIPFrame,
)
from xknx.io.gateway_scanner import GatewayDescriptor
from xknx.knxip import (
    HPAI,
    ConnectionStateRequest,
    ConnectionStateResponse,
    ConnectRequest,
    ConnectRequestType,
    ConnectResponse,
    ConnectResponseData,
    DIBDeviceInformation,
    DIBGeneric,
    DIBServiceFamily,
    DIBSuppSVCFamilies,
    DisconnectRequest,
    DisconnectResponse,
    HostProtocol,
    KNXIPFrame,
    KNXIPHeader,
    KNXMedium,
    RoutingIndication,
    SearchRequest,
    SearchResponse,
    SearchResponseExtended,
    TunnellingAck,
    TunnellingFeatureGet,
    TunnellingFeatureInfo,
    TunnellingFeatureResponse,
    TunnellingFeatureSet,
    TunnellingFeatureType,
    TunnellingRequest,
)
from xknx.profile.const import ResourceKNXNETIPPropertyId, ResourceObjectType
from xknx.remote_value import RemoteValue
from xknx.telegram import GroupAddress, IndividualAddress, Telegram, TelegramDirection
from xknx.telegram.apci import GroupValueWrite


class TestStringRepresentations:
    """Test class for Configuration logic."""

    @patch.multiple(RemoteValue, __abstractmethods__=set())
    def test_remote_value(self) -> None:
        """Test string representation of remote value."""
        xknx = XKNX()
        remote_value: RemoteValue[Any] = RemoteValue(
            xknx,
            group_address="1/2/3",
            device_name="MyDevice",
            group_address_state="1/2/4",
        )
        assert (
            str(remote_value)
            == '<RemoteValue device_name="MyDevice" feature_name="Unknown" <1/2/3, 1/2/4, [], None /> />'
        )

        remote_value.to_knx = lambda value: value  # to bypass NotImplementedError
        remote_value.value = 34
        assert (
            str(remote_value)
            == '<RemoteValue device_name="MyDevice" feature_name="Unknown" '
            "<1/2/3, 1/2/4, [], 34 /> />"
        )

        remote_value_passive: RemoteValue[Any] = RemoteValue(
            xknx,
            group_address=["1/2/3", "1/2/4", "i-test"],
            device_name="MyDevice",
        )
        assert (
            str(remote_value_passive)
            == "<RemoteValue device_name=\"MyDevice\" feature_name=\"Unknown\" <1/2/3, None, ['1/2/4', 'i-test'], None /> />"
        )

    def test_binary_sensor(self) -> None:
        """Test string representation of binary sensor object."""
        xknx = XKNX()
        binary_sensor = BinarySensor(xknx, name="Fnord", group_address_state="1/2/3")
        assert (
            str(binary_sensor)
            == '<BinarySensor name="Fnord" remote_value=<None, 1/2/3, [], None /> state=None />'
        )

    def test_climate(self) -> None:
        """Test string representation of climate object."""
        xknx = XKNX()
        climate = Climate(
            xknx,
            name="Wohnzimmer",
            group_address_temperature="1/2/1",
            group_address_target_temperature="1/2/2",
            group_address_setpoint_shift="1/2/3",
            group_address_setpoint_shift_state="1/2/4",
            temperature_step=0.1,
            setpoint_shift_max=20,
            setpoint_shift_min=-20,
            group_address_on_off="1/2/14",
            group_address_on_off_state="1/2/15",
            group_address_fan_speed="1/2/16",
            group_address_fan_speed_state="1/2/17",
            group_address_swing="1/2/18",
            group_address_swing_state="1/2/19",
            group_address_horizontal_swing="1/2/20",
            group_address_horizontal_swing_state="1/2/21",
        )
        assert (
            str(climate) == '<Climate name="Wohnzimmer" '
            "temperature=<None, 1/2/1, [], None /> "
            "target_temperature=<1/2/2, None, [], None /> "
            'temperature_step="0.1" '
            "setpoint_shift=<1/2/3, 1/2/4, [], None /> "
            'setpoint_shift_max="20" setpoint_shift_min="-20" '
            "group_address_on_off=<1/2/14, 1/2/15, [], None /> "
            "group_address_fan_speed=<1/2/16, 1/2/17, [], None /> "
            "group_address_swing=<1/2/18, 1/2/19, [], None /> "
            "group_address_horizontal_swing=<1/2/20, 1/2/21, [], None /> />"
        )

    def test_climate_mode(self) -> None:
        """Test string representation of climate mode object."""
        xknx = XKNX()
        climate_mode = ClimateMode(
            xknx,
            name="Wohnzimmer Mode",
            group_address_operation_mode="1/2/5",
            group_address_operation_mode_state="1/2/6",
            group_address_operation_mode_protection="1/2/7",
            group_address_operation_mode_economy="1/2/8",
            group_address_operation_mode_comfort="1/2/9",
            group_address_controller_status="1/2/10",
            group_address_controller_status_state="1/2/11",
            group_address_controller_mode="1/2/12",
            group_address_controller_mode_state="1/2/13",
        )
        assert (
            str(climate_mode) == '<ClimateMode name="Wohnzimmer Mode" '
            "operation_mode=<1/2/5, 1/2/6, [], None /> "
            "controller_mode=<1/2/12, 1/2/13, [], None /> "
            "controller_status=<1/2/10, 1/2/11, [], None /> />"
        )

    def test_cover(self) -> None:
        """Test string representation of cover object."""
        xknx = XKNX()
        cover = Cover(
            xknx,
            name="Rolladen",
            group_address_long="1/2/2",
            group_address_short="1/2/3",
            group_address_stop="1/2/4",
            group_address_position="1/2/5",
            group_address_position_state="1/2/6",
            group_address_angle="1/2/7",
            group_address_angle_state="1/2/8",
            group_address_locked_state="1/2/9",
            travel_time_down=8,
            travel_time_up=10,
        )
        assert (
            str(cover) == '<Cover name="Rolladen" '
            "updown=<1/2/2, None, [], None /> "
            "step=<1/2/3, None, [], None /> "
            "stop_=<1/2/4, None, [], None /> "
            "position_current=<None, 1/2/6, [], None /> "
            "position_target=<1/2/5, None, [], None /> "
            "angle=<1/2/7, 1/2/8, [], None /> "
            "locked=<None, 1/2/9, [], None /> "
            'travel_time_down="8" '
            'travel_time_up="10" />'
        )

    def test_fan(self) -> None:
        """Test string representation of fan object."""
        xknx = XKNX()
        fan = Fan(
            xknx,
            name="Dunstabzug",
            group_address_speed="1/2/3",
            group_address_speed_state="1/2/4",
            group_address_oscillation="1/2/5",
            group_address_oscillation_state="1/2/6",
        )
        assert (
            str(fan)
            == '<Fan name="Dunstabzug" speed=<1/2/3, 1/2/4, [], None /> oscillation=<1/2/5, 1/2/6, [], None /> />'
        )

    def test_light(self) -> None:
        """Test string representation of non dimmable light object."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="Licht",
            group_address_switch="1/2/3",
            group_address_switch_state="1/2/4",
        )
        assert str(light) == '<Light name="Licht" switch=<1/2/3, 1/2/4, [], None /> />'

    def test_light_dimmable(self) -> None:
        """Test string representation of dimmable light object."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="Licht",
            group_address_switch="1/2/3",
            group_address_switch_state="1/2/4",
            group_address_brightness="1/2/5",
            group_address_brightness_state="1/2/6",
        )
        assert (
            str(light) == '<Light name="Licht" '
            "switch=<1/2/3, 1/2/4, [], None /> "
            "brightness=<1/2/5, 1/2/6, [], None /> />"
        )

    def test_light_color(self) -> None:
        """Test string representation of dimmable light object."""
        xknx = XKNX()
        light = Light(
            xknx,
            name="Licht",
            group_address_switch="1/2/3",
            group_address_switch_state="1/2/4",
            group_address_color="1/2/5",
            group_address_color_state="1/2/6",
        )
        assert (
            str(light) == '<Light name="Licht" '
            "switch=<1/2/3, 1/2/4, [], None /> "
            "color=<1/2/5, 1/2/6, [], None /> />"
        )

    async def test_notification(self) -> None:
        """Test string representation of notification object."""
        xknx = XKNX()
        notification = Notification(
            xknx, name="Alarm", group_address="1/2/3", group_address_state="1/2/4"
        )
        assert (
            str(notification)
            == '<Notification name="Alarm" message=<1/2/3, 1/2/4, [], None /> />'
        )
        await notification.set("Einbrecher im Haus")
        notification.process(xknx.telegrams.get_nowait())
        assert (
            str(notification) == '<Notification name="Alarm" '
            "message=<1/2/3, 1/2/4, [], 'Einbrecher im ' /> />"
        )

    def test_scene(self) -> None:
        """Test string representation of scene object."""
        xknx = XKNX()
        scene = Scene(xknx, name="Romantic", group_address="1/2/3", scene_number=23)
        assert (
            str(scene)
            == '<Scene name="Romantic" scene_value=<1/2/3, None, [], None /> scene_number="23" />'
        )

    async def test_sensor(self) -> None:
        """Test string representation of sensor object."""
        xknx = XKNX()
        sensor = Sensor(
            xknx, name="MeinSensor", group_address_state="1/2/3", value_type="percent"
        )
        assert (
            str(sensor)
            == '<Sensor name="MeinSensor" sensor=<None, 1/2/3, [], None /> value=None unit="%"/>'
        )
        # self.loop.run_until_complete(sensor.sensor_value.set(25))
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTArray(0x40)),
        )
        sensor.process_group_write(telegram)
        assert (
            str(sensor)
            == '<Sensor name="MeinSensor" sensor=<None, 1/2/3, [], 25 /> value=25 unit="%"/>'
        )

    async def test_expose_sensor(self) -> None:
        """Test string representation of expose sensor object."""
        xknx = XKNX()
        sensor = ExposeSensor(
            xknx, name="MeinSensor", group_address="1/2/3", value_type="percent"
        )
        assert (
            str(sensor)
            == '<ExposeSensor name="MeinSensor" sensor=<1/2/3, None, [], None /> value=None unit="%"/>'
        )
        await sensor.set(25)
        sensor.process(xknx.telegrams.get_nowait())
        assert (
            str(sensor)
            == '<ExposeSensor name="MeinSensor" sensor=<1/2/3, None, [], 25 /> value=25 unit="%"/>'
        )

    def test_switch(self) -> None:
        """Test string representation of switch object."""
        xknx = XKNX()
        switch = Switch(
            xknx, name="Schalter", group_address="1/2/3", group_address_state="1/2/4"
        )
        assert (
            str(switch)
            == '<Switch name="Schalter" switch=<1/2/3, 1/2/4, [], None /> />'
        )

    async def test_weather(self) -> None:
        """Test string representation of switch object."""
        xknx = XKNX()
        weather = Weather(
            xknx,
            "Home",
            group_address_temperature="7/0/1",
            group_address_brightness_south="7/0/5",
            group_address_brightness_east="7/0/4",
            group_address_brightness_west="7/0/3",
            group_address_wind_speed="7/0/2",
            group_address_wind_bearing="7/0/6",
            group_address_day_night="7/0/7",
            group_address_rain_alarm="7/0/0",
            group_address_frost_alarm="7/0/8",
            group_address_air_pressure="7/0/9",
            group_address_humidity="7/0/9",
            group_address_wind_alarm="7/0/10",
        )
        assert (
            str(weather)
            == '<Weather name="Home" temperature=<None, 7/0/1, [], None /> '
            "brightness_south=<None, 7/0/5, [], None /> brightness_north=<None, None, [], None /> "
            "brightness_west=<None, 7/0/3, [], None /> "
            "brightness_east=<None, 7/0/4, [], None /> wind_speed=<None, 7/0/2, [], None /> "
            "wind_bearing=<None, 7/0/6, [], None /> rain_alarm=<None, 7/0/0, [], None /> "
            "wind_alarm=<None, 7/0/10, [], None /> frost_alarm=<None, 7/0/8, [], None /> "
            "day_night=<None, 7/0/7, [], None /> air_pressure=<None, 7/0/9, [], None /> "
            "humidity=<None, 7/0/9, [], None /> />"
        )

        telegram = Telegram(
            destination_address=GroupAddress("7/0/10"),
            direction=TelegramDirection.INCOMING,
            payload=GroupValueWrite(DPTBinary(1)),
        )
        weather.process_group_write(telegram)

        assert (
            str(weather)
            == '<Weather name="Home" temperature=<None, 7/0/1, [], None /> '
            "brightness_south=<None, 7/0/5, [], None /> brightness_north=<None, None, [], None /> "
            "brightness_west=<None, 7/0/3, [], None /> "
            "brightness_east=<None, 7/0/4, [], None /> wind_speed=<None, 7/0/2, [], None /> "
            "wind_bearing=<None, 7/0/6, [], None /> rain_alarm=<None, 7/0/0, [], None /> "
            "wind_alarm=<None, 7/0/10, [], True /> "
            "frost_alarm=<None, 7/0/8, [], None /> day_night=<None, 7/0/7, [], None /> "
            "air_pressure=<None, 7/0/9, [], None /> humidity=<None, 7/0/9, [], None /> />"
        )

    def test_datetime(self) -> None:
        """Test string representation of datetime object."""
        xknx = XKNX()
        date_time = DateTimeDevice(
            xknx, name="Zeit", group_address="1/2/3", localtime=False
        )
        assert (
            str(date_time)
            == '<DateTimeDevice name="Zeit" remote_value=<1/2/3, None, [], None /> />'
        )

    def test_could_not_parse_telegramn_exception(self) -> None:
        """Test string representation of CouldNotParseTelegram exception."""
        exception = CouldNotParseTelegram(description="Fnord")
        assert str(exception) == '<CouldNotParseTelegram description="Fnord" />'

    def test_could_not_parse_telegramn_exception_parameter(self) -> None:
        """Test string representation of CouldNotParseTelegram exception."""
        exception = CouldNotParseTelegram(description="Fnord", one="one", two="two")
        assert (
            str(exception)
            == "<CouldNotParseTelegram description=\"Fnord\" one='one' two='two'/>"
        )

    def test_could_not_parse_knxip_exception(self) -> None:
        """Test string representation of CouldNotParseKNXIP exception."""
        exception = CouldNotParseKNXIP(description="Fnord")
        assert str(exception) == '<CouldNotParseKNXIP description="Fnord" />'

    def test_conversion_error_exception(self) -> None:
        """Test string representation of ConversionError exception."""
        exception = ConversionError(description="Fnord")
        assert str(exception) == '<ConversionError description="Fnord" />'

    def test_conversion_error_exception_parameter(self) -> None:
        """Test string representation of ConversionError exception."""
        exception = ConversionError(description="Fnord", one="one", two="two")
        assert (
            str(exception)
            == "<ConversionError description=\"Fnord\" one='one' two='two'/>"
        )

    def test_could_not_parse_address_exception(self) -> None:
        """Test string representation of CouldNotParseAddress exception."""
        exception = CouldNotParseAddress(address="1/2/1000")
        assert str(exception) == "<CouldNotParseAddress address='1/2/1000' />"

    def test_device_illegal_value_exception(self) -> None:
        """Test string representation of DeviceIllegalValue exception."""
        exception = DeviceIllegalValue(value=12, description="Fnord exceeded")
        assert (
            str(exception)
            == '<DeviceIllegalValue description="Fnord exceeded" value=12 />'
        )

    def test_incomplete_knxip_frame_excetpion(self) -> None:
        """Test string representation of IncompleteKNXIPFrame exception."""
        exception = IncompleteKNXIPFrame("Hello")
        assert str(exception) == '<IncompleteKNXIPFrame description="Hello" />'

    def test_address(self) -> None:
        """Test string representation of address object."""
        address = GroupAddress("1/2/3")
        assert repr(address) == 'GroupAddress("1/2/3")'
        assert str(address) == "1/2/3"

    def test_dpt_array(self) -> None:
        """Test string representation of DPTBinary."""
        dpt_array = DPTArray([0x01, 0x02])
        assert str(dpt_array) == '<DPTArray value="[0x1,0x2]" />'

    def test_dpt_binary(self) -> None:
        """Test string representation of DPTBinary."""
        dpt_binary = DPTBinary(7)
        assert str(dpt_binary) == '<DPTBinary value="7" />'

    def test_telegram(self) -> None:
        """Test string representation of Telegram."""
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(7)),
        )
        assert (
            str(telegram) == '<Telegram direction="Outgoing" source_address="0.0.0" '
            'destination_address="1/2/3" payload="<GroupValueWrite value="<DPTBinary value="7" />" />" />'
        )

    def test_dib_generic(self) -> None:
        """Test string representation of DIBGeneric."""
        dib = DIBGeneric()
        dib.dtc = 0x01
        dib.data = [0x02, 0x03, 0x04]
        assert str(dib) == '<DIB dtc="1" data="0x02, 0x03, 0x04" />'

    def test_dib_supp_svc_families(self) -> None:
        """Test string representation of DIBSuppSVCFamilies."""
        dib = DIBSuppSVCFamilies()
        dib.families.append(DIBSuppSVCFamilies.Family(DIBServiceFamily.CORE, "1"))
        dib.families.append(
            DIBSuppSVCFamilies.Family(DIBServiceFamily.DEVICE_MANAGEMENT, "2")
        )
        assert (
            str(dib)
            == '<DIBSuppSVCFamilies families="[DIBServiceFamily.CORE version: 1, DIBServiceFamily.DEVICE_MANAGEMENT version: 2]" />'
        )

    def test_dib_device_informatio(self) -> None:
        """Test string representation of DIBDeviceInformation."""
        dib = DIBDeviceInformation()
        dib.knx_medium = KNXMedium.TP1
        dib.programming_mode = False
        dib.individual_address = IndividualAddress("1.1.0")
        dib.name = "Gira KNX/IP-Router"
        dib.mac_address = "00:01:02:03:04:05"
        dib.multicast_address = "224.0.23.12"
        dib.serial_number = "13:37:13:37:13:37"
        dib.project_number = 564
        dib.installation_number = 2
        assert (
            str(dib) == "<DIBDeviceInformation \n"
            '\tknx_medium="KNXMedium.TP1" \n'
            '\tprogramming_mode="False" \n'
            '\tindividual_address="1.1.0" \n'
            '\tinstallation_number="2" \n'
            '\tproject_number="564" \n'
            '\tserial_number="13:37:13:37:13:37" \n'
            '\tmulticast_address="224.0.23.12" \n'
            '\tmac_address="00:01:02:03:04:05" \n'
            '\tname="Gira KNX/IP-Router" />'
        )

    def test_hpai(self) -> None:
        """Test string representation of HPAI."""
        hpai_udp = HPAI(ip_addr="192.168.42.1", port=33941)
        assert str(hpai_udp) == "192.168.42.1:33941/udp"
        assert repr(hpai_udp) == "HPAI('192.168.42.1', 33941, HostProtocol.IPV4_UDP)"

        hpai_tcp = HPAI(ip_addr="10.1.4.1", port=3671, protocol=HostProtocol.IPV4_TCP)
        assert str(hpai_tcp) == "10.1.4.1:3671/tcp"
        assert repr(hpai_tcp) == "HPAI('10.1.4.1', 3671, HostProtocol.IPV4_TCP)"

    def test_header(self) -> None:
        """Test string representation of KNX/IP-Header."""
        header = KNXIPHeader()
        header.total_length = 42
        assert (
            str(header)
            == '<KNXIPHeader HeaderLength="6" ProtocolVersion="16" KNXIPServiceType="ROUTING_INDICATION" TotalLength="42" '
            "/>"
        )

    def test_connect_request(self) -> None:
        """Test string representation of KNX/IP ConnectRequest."""
        connect_request = ConnectRequest()
        connect_request.request_type = ConnectRequestType.TUNNEL_CONNECTION
        connect_request.control_endpoint = HPAI(ip_addr="192.168.42.1", port=33941)
        connect_request.data_endpoint = HPAI(ip_addr="192.168.42.2", port=33942)
        assert (
            str(connect_request)
            == '<ConnectRequest control_endpoint="192.168.42.1:33941/udp" data_endpoint="192.168.42.2:33942/udp" '
            'cri="<ConnectRequestInformation connection_type="TUNNEL_CONNECTION" knx_layer="DATA_LINK_LAYER" />" />'
        )

    def test_connect_response(self) -> None:
        """Test string representatoin of KNX/IP ConnectResponse."""
        connect_response = ConnectResponse()
        connect_response.communication_channel = 13
        connect_response.data_endpoint = HPAI(ip_addr="192.168.42.1", port=33941)
        connect_response.crd = ConnectResponseData(
            request_type=ConnectRequestType.TUNNEL_CONNECTION,
            individual_address=IndividualAddress("1.2.3"),
        )
        assert (
            str(connect_response)
            == '<ConnectResponse communication_channel="13" status_code="ErrorCode.E_NO_ERROR" '
            'data_endpoint="192.168.42.1:33941/udp" '
            'crd="<ConnectResponseData request_type="ConnectRequestType.TUNNEL_CONNECTION" individual_address="1.2.3" />" />'
        )

    def test_disconnect_request(self) -> None:
        """Test string representation of KNX/IP DisconnectRequest."""
        disconnect_request = DisconnectRequest()
        disconnect_request.communication_channel_id = 13
        disconnect_request.control_endpoint = HPAI(ip_addr="192.168.42.1", port=33941)
        assert (
            str(disconnect_request)
            == '<DisconnectRequest communication_channel_id="13" control_endpoint="192.168.42.1:33941/udp" />'
        )

    def test_disconnect_response(self) -> None:
        """Test string representation of KNX/IP DisconnectResponse."""
        disconnect_response = DisconnectResponse()
        disconnect_response.communication_channel_id = 23
        assert (
            str(disconnect_response)
            == '<DisconnectResponse communication_channel_id="23" status_code="ErrorCode.E_NO_ERROR" />'
        )

    def test_connectionstate_request(self) -> None:
        """Test string representation of KNX/IP ConnectionStateRequest."""
        connectionstate_request = ConnectionStateRequest()
        connectionstate_request.communication_channel_id = 23
        connectionstate_request.control_endpoint = HPAI(
            ip_addr="192.168.42.1", port=33941
        )
        assert (
            str(connectionstate_request)
            == '<ConnectionStateRequest communication_channel_id="23", control_endpoint="192.168.42.1:33941/udp" />'
        )

    def test_connectionstate_response(self) -> None:
        """Test string representation of KNX/IP ConnectionStateResponse."""
        connectionstate_response = ConnectionStateResponse()
        connectionstate_response.communication_channel_id = 23
        assert (
            str(connectionstate_response)
            == '<ConnectionStateResponse communication_channel_id="23" status_code="ErrorCode.E_NO_ERROR" />'
        )

    def test_search_reqeust(self) -> None:
        """Test string representation of KNX/IP SearchRequest."""
        search_request = SearchRequest()
        assert (
            str(search_request)
            == '<SearchRequest discovery_endpoint="0.0.0.0:0/udp" />'
        )

    def test_search_response(self) -> None:
        """Test string representation of KNX/IP SearchResponse."""
        search_response = SearchResponse()
        search_response.control_endpoint = HPAI(ip_addr="192.168.42.1", port=33941)
        search_response.dibs.append(DIBGeneric())
        search_response.dibs.append(DIBGeneric())
        assert (
            str(search_response)
            == '<SearchResponse control_endpoint="192.168.42.1:33941/udp" dibs="[\n'
            '<DIB dtc="0" data="" />,\n'
            '<DIB dtc="0" data="" />\n'
            ']" />'
        )

    def test_search_response_extended(self) -> None:
        """Test string representation of KNX/IP SearchResponseExtended."""
        search_response = SearchResponseExtended()
        search_response.control_endpoint = HPAI(ip_addr="192.168.42.1", port=33941)
        search_response.dibs.append(DIBGeneric())
        search_response.dibs.append(DIBGeneric())
        assert (
            str(search_response)
            == '<SearchResponseExtended control_endpoint="192.168.42.1:33941/udp" dibs="[\n'
            '<DIB dtc="0" data="" />,\n'
            '<DIB dtc="0" data="" />\n'
            ']" />'
        )

    def test_tunnelling_request(self) -> None:
        """Test string representation of KNX/IP TunnellingRequest."""
        tunnelling_request = TunnellingRequest()
        tunnelling_request.communication_channel_id = 23
        tunnelling_request.sequence_counter = 42
        assert (
            str(tunnelling_request)
            == '<TunnellingRequest communication_channel_id="23" sequence_counter="42" cemi="" />'
        )

    def test_tunnelling_ack(self) -> None:
        """Test string representation of KNX/IP TunnellingAck."""
        tunnelling_ack = TunnellingAck()
        tunnelling_ack.communication_channel_id = 23
        tunnelling_ack.sequence_counter = 42
        assert (
            str(tunnelling_ack)
            == '<TunnellingAck communication_channel_id="23" sequence_counter="42" status_code="ErrorCode.E_NO_ERROR" />'
        )

    def test_tunnelling_feature_get(self) -> None:
        """Test string representation of KNX/IP TunnellingFeatureGet."""
        tunnelling_feature = TunnellingFeatureGet()
        tunnelling_feature.communication_channel_id = 23
        tunnelling_feature.sequence_counter = 42
        tunnelling_feature.feature_type = TunnellingFeatureType.BUS_CONNECTION_STATUS
        assert (
            str(tunnelling_feature)
            == '<TunnellingFeatureGet communication_channel_id="23" sequence_counter="42" '
            'feature_type="TunnellingFeatureType.BUS_CONNECTION_STATUS" />'
        )

    def test_tunnelling_feature_info(self) -> None:
        """Test string representation of KNX/IP TunnellingFeatureInfo."""
        tunnelling_feature = TunnellingFeatureInfo()
        tunnelling_feature.communication_channel_id = 23
        tunnelling_feature.sequence_counter = 42
        tunnelling_feature.feature_type = TunnellingFeatureType.BUS_CONNECTION_STATUS
        tunnelling_feature.data = b"\x01\x00"
        assert (
            str(tunnelling_feature)
            == '<TunnellingFeatureInfo communication_channel_id="23" sequence_counter="42" '
            'feature_type="TunnellingFeatureType.BUS_CONNECTION_STATUS" data="0100" />'
        )

    def test_tunnelling_feature_response(self) -> None:
        """Test string representation of KNX/IP TunnellingFeatureResponse."""
        tunnelling_feature = TunnellingFeatureResponse()
        tunnelling_feature.communication_channel_id = 23
        tunnelling_feature.sequence_counter = 42
        tunnelling_feature.feature_type = TunnellingFeatureType.BUS_CONNECTION_STATUS
        tunnelling_feature.data = b"\x01\x00"
        assert (
            str(tunnelling_feature)
            == '<TunnellingFeatureResponse communication_channel_id="23" sequence_counter="42" status_code="ErrorCode.E_NO_ERROR" '
            'feature_type="TunnellingFeatureType.BUS_CONNECTION_STATUS" return_code="ReturnCode.E_SUCCESS" data="0100" />'
        )

    def test_tunnelling_feature_set(self) -> None:
        """Test string representation of KNX/IP TunnellingFeatureSet."""
        tunnelling_feature = TunnellingFeatureSet()
        tunnelling_feature.communication_channel_id = 23
        tunnelling_feature.sequence_counter = 42
        tunnelling_feature.feature_type = (
            TunnellingFeatureType.INTERFACE_FEATURE_INFO_ENABLE
        )
        tunnelling_feature.data = b"\x01\x00"
        assert (
            str(tunnelling_feature)
            == '<TunnellingFeatureSet communication_channel_id="23" sequence_counter="42" '
            'feature_type="TunnellingFeatureType.INTERFACE_FEATURE_INFO_ENABLE" data="0100" />'
        )

    def test_cemi_ldata_frame(self) -> None:
        """Test string representation of KNX/IP CEMI Frame."""
        cemi_frame = CEMIFrame(
            code=CEMIMessageCode.L_DATA_IND,
            data=CEMILData.init_from_telegram(
                telegram=Telegram(
                    destination_address=GroupAddress("1/2/5"),
                    payload=GroupValueWrite(DPTBinary(7)),
                ),
                src_addr=IndividualAddress("1.2.3"),
            ),
        )
        assert (
            str(cemi_frame)
            == '<CEMIFrame code="L_DATA_IND" info="CEMIInfo("")" data="CEMILData(src_addr="IndividualAddress("1.2.3")" '
            'dst_addr="GroupAddress("1/2/5")" flags="1011110011100000" tpci="TDataGroup()" '
            'payload="<GroupValueWrite value="<DPTBinary value="7" />" />")" />'
        )

    def test_cemi_mprop_frame(self) -> None:
        """Test string representation of KNX/IP CEMI Frame."""
        cemi_frame = CEMIFrame(
            code=CEMIMessageCode.M_PROP_READ_REQ,
            data=CEMIMPropReadResponse(
                property_info=CEMIMPropInfo(
                    object_type=ResourceObjectType.OBJECT_KNXNETIP_PARAMETER,
                    property_id=ResourceKNXNETIPPropertyId.PID_KNX_INDIVIDUAL_ADDRESS,
                ),
                data=IndividualAddress("1.2.3").to_knx(),
            ),
        )
        assert (
            str(cemi_frame)
            == '<CEMIFrame code="M_PROP_READ_REQ" data="CEMIMPropReadResponse(object_type="11" object_instance="1" '
            'property_id="52" number_of_elements="1" start_index="1" error_code="None" data="1203" )" />'
        )

    def test_knxip_frame(self) -> None:
        """Test string representation of KNX/IP Frame."""
        search_request = SearchRequest()
        knxipframe = KNXIPFrame.init_from_body(search_request)
        assert (
            str(knxipframe)
            == '<KNXIPFrame <KNXIPHeader HeaderLength="6" ProtocolVersion="16" KNXIPServiceType="SEARCH_REQUEST" '
            'TotalLength="14" /> body="<SearchRequest discovery_endpoint="0.0.0.0:0/udp" />" />'
        )

    #
    # Gateway Scanner
    #
    def test_gateway_descriptor(self) -> None:
        """Test string representation of GatewayDescriptor."""
        gateway_descriptor = GatewayDescriptor(
            name="KNX-Interface",
            ip_addr="192.168.2.3",
            port=1234,
            local_interface="en1",
            local_ip="192.168.2.50",
            supports_tunnelling=True,
            supports_routing=False,
            individual_address=IndividualAddress("1.1.1"),
        )
        assert str(gateway_descriptor) == "1.1.1 - KNX-Interface @ 192.168.2.3:1234"

    #
    # Routing Indication
    #
    def test_routing_indication_str(self) -> None:
        """Test string representation of GatewayDescriptor."""
        routing_indication = RoutingIndication()
        assert str(routing_indication) == '<RoutingIndication cemi="" />'



"""Implementation of KNX DPT 17 Scene."""

from __future__ import annotations

from xknx.exceptions import ConversionError

from .dpt_5 import DPTValue1ByteUnsigned
from .payload import DPTArray, DPTBinary


class DPTSceneNumber(DPTValue1ByteUnsigned):
    """
    Abstraction for KNX 1 Octet Scene Number.

    DPT 17.001
    """

    dpt_main_number = 17
    dpt_sub_number = 1
    value_type = "scene_number"

    value_min = 1
    value_max = 64

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        value = cls.validate_payload(payload)[0] + 1

        if not cls._test_boundaries(value):
            raise ConversionError(
                f"Could not parse {cls.dpt_name()}", value=value, payload=payload
            )

        return value

    @classmethod
    def to_knx(cls, value: int | float) -> DPTArray:
        """Serialize to KNX/IP raw data."""
        try:
            knx_value = int(value) - 1
            if not cls._test_boundaries(knx_value + 1):
                raise ValueError("Value out of range")
            return DPTArray(knx_value)
        except ValueError as err:
            raise ConversionError(
                f"Could not serialize {cls.dpt_name()}", value=value
            ) from err



"""Unit test for KNX/IP TunnellingRequest objects."""

import pytest

from xknx.cemi import CEMIFrame, CEMILData, CEMIMessageCode
from xknx.dpt import DPTBinary
from xknx.exceptions import CouldNotParseKNXIP
from xknx.knxip import KNXIPFrame, TunnellingRequest
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueWrite


class TestKNXIPTunnellingRequest:
    """Test class for KNX/IP TunnellingRequest objects."""

    def test_connect_request(self) -> None:
        """Test parsing and streaming connection tunneling request KNX/IP packet."""
        raw = bytes.fromhex(
            "06 10 04 20 00 15 04 01 17 00 11 00 BC E0 00 00 48 08 01 00 81"
        )
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, TunnellingRequest)
        assert knxipframe.body.communication_channel_id == 1
        assert knxipframe.body.sequence_counter == 23
        assert isinstance(knxipframe.body.raw_cemi, bytes)

        incoming_cemi = CEMIFrame.from_knx(knxipframe.body.raw_cemi)
        assert incoming_cemi.data.telegram() == Telegram(
            destination_address=GroupAddress("9/0/8"),
            payload=GroupValueWrite(DPTBinary(1)),
        )

        outgoing_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_REQ,
            data=CEMILData.init_from_telegram(
                Telegram(
                    destination_address=GroupAddress("9/0/8"),
                    payload=GroupValueWrite(DPTBinary(1)),
                ),
            ),
        )
        tunnelling_request = TunnellingRequest(
            communication_channel_id=1,
            sequence_counter=23,
            raw_cemi=outgoing_cemi.to_knx(),
        )
        knxipframe2 = KNXIPFrame.init_from_body(tunnelling_request)

        assert knxipframe2.to_knx() == raw

    def test_from_knx_wrong_header(self) -> None:
        """Test parsing and streaming wrong TunnellingRequest (wrong header length byte)."""
        raw = bytes((0x06, 0x10, 0x04, 0x20, 0x00, 0x15, 0x03))
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_from_knx_wrong_header2(self) -> None:
        """Test parsing and streaming wrong TunnellingRequest (wrong header length)."""
        raw = bytes((0x06, 0x10, 0x04, 0x20, 0x00, 0x15, 0x04))
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)



"""
Module for Serialization and Deserialization of a KNX Connectionstate Response information.

Connectionstate requests are used to determine if a tunnel connection is still active and valid.
With a connectionstate response the receiving party acknowledges the valid processing of the request.
"""

from __future__ import annotations

from xknx.exceptions import CouldNotParseKNXIP

from .body import KNXIPBodyResponse
from .error_code import ErrorCode
from .knxip_enum import KNXIPServiceType


class ConnectionStateResponse(KNXIPBodyResponse):
    """Representation of a KNX Connection State Response."""

    SERVICE_TYPE = KNXIPServiceType.CONNECTIONSTATE_RESPONSE
    LENGTH = 2

    def __init__(
        self,
        communication_channel_id: int = 1,
        status_code: ErrorCode = ErrorCode.E_NO_ERROR,
    ) -> None:
        """Initialize ConnectionStateResponse object."""
        self.communication_channel_id = communication_channel_id
        self.status_code = status_code

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return ConnectionStateResponse.LENGTH

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) < ConnectionStateResponse.LENGTH:
            raise CouldNotParseKNXIP("ConnectionStateResponse info has wrong length")
        self.communication_channel_id = raw[0]
        self.status_code = ErrorCode(raw[1])
        return ConnectionStateResponse.LENGTH

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return bytes((self.communication_channel_id, self.status_code.value))

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            "<ConnectionStateResponse "
            f'communication_channel_id="{self.communication_channel_id}" '
            f'status_code="{self.status_code}" />'
        )



"""Unit test for RawValue objects."""

from unittest.mock import Mock

import pytest

from xknx import XKNX
from xknx.devices import RawValue
from xknx.dpt import DPTArray, DPTBinary
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueRead, GroupValueResponse, GroupValueWrite


class TestRawValue:
    """Test class for RawValue objects."""

    @pytest.mark.parametrize(
        "payload_length,raw_payload,expected_state",
        [
            # DPT-14 values are according to ETS group monitor values
            (
                0,
                DPTBinary(0),
                0,
            ),
            (
                0,
                DPTBinary(True),
                1,
            ),
            (
                0,
                DPTBinary(45),
                45,
            ),
            (
                1,
                DPTArray((0x6B,)),
                107,
            ),
            (
                1,
                DPTArray((0xFC,)),
                252,
            ),
            (
                2,
                DPTArray((0x6C, 0x95)),
                27797,
            ),
        ],
    )
    async def test_payloads(
        self,
        payload_length: int,
        raw_payload: DPTArray | DPTBinary,
        expected_state: int,
    ) -> None:
        """Test raw value types."""
        xknx = XKNX()
        raw_value = RawValue(
            xknx,
            "Test",
            payload_length=payload_length,
            group_address="1/2/3",
        )
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(value=raw_payload),
        )
        raw_value.process(telegram)
        assert raw_value.resolve_state() == expected_state
        assert raw_value.last_telegram == telegram

    #
    # TEST RESPOND
    #
    async def test_respond_to_read(self) -> None:
        """Test respond_to_read function."""
        xknx = XKNX()
        responding = RawValue(
            xknx,
            "TestSensor1",
            2,
            group_address="1/1/1",
            respond_to_read=True,
        )
        non_responding = RawValue(
            xknx,
            "TestSensor2",
            2,
            group_address="1/1/1",
            respond_to_read=False,
        )
        responding_multiple = RawValue(
            xknx,
            "TestSensor3",
            2,
            group_address=["1/1/1", "3/3/3"],
            group_address_state="2/2/2",
            respond_to_read=True,
        )
        #  set initial payload of RawValue
        responding.remote_value.value = 256
        non_responding.remote_value.value = 256
        responding_multiple.remote_value.value = 256

        read_telegram = Telegram(
            destination_address=GroupAddress("1/1/1"), payload=GroupValueRead()
        )
        # verify no response when respond is False
        non_responding.process(read_telegram)
        assert xknx.telegrams.qsize() == 0

        # verify response when respond is True
        responding.process(read_telegram)
        assert xknx.telegrams.qsize() == 1
        response = xknx.telegrams.get_nowait()
        assert response == Telegram(
            destination_address=GroupAddress("1/1/1"),
            payload=GroupValueResponse(DPTArray((0x01, 0x00))),
        )
        # verify no response when GroupValueRead request is not for group_address
        responding_multiple.process(read_telegram)
        assert xknx.telegrams.qsize() == 1
        response = xknx.telegrams.get_nowait()
        assert response == Telegram(
            destination_address=GroupAddress("1/1/1"),
            payload=GroupValueResponse(DPTArray((0x01, 0x00))),
        )
        responding_multiple.process(
            Telegram(
                destination_address=GroupAddress("2/2/2"), payload=GroupValueRead()
            )
        )
        responding_multiple.process(
            Telegram(
                destination_address=GroupAddress("3/3/3"), payload=GroupValueRead()
            )
        )
        assert xknx.telegrams.qsize() == 0

    #
    # TEST PROCESS CALLBACK
    #

    async def test_process_callback(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if callback is called."""

        xknx = XKNX()
        sensor = RawValue(
            xknx,
            "TestSensor",
            2,
            group_address="1/2/3",
        )
        after_update_callback = Mock()
        sensor.register_device_updated_cb(after_update_callback)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x01, 0x02))),
        )
        sensor.process(telegram)
        after_update_callback.assert_called_with(sensor)
        assert sensor.last_telegram == telegram
        # consecutive telegrams with same payload shall only trigger one callback
        after_update_callback.reset_mock()
        sensor.process(telegram)
        after_update_callback.assert_not_called()

    async def test_process_callback_always(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if callback is called."""

        xknx = XKNX()
        sensor = RawValue(
            xknx,
            "TestSensor",
            2,
            group_address="1/2/3",
            always_callback=True,
        )
        after_update_callback = Mock()
        sensor.register_device_updated_cb(after_update_callback)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x01, 0x02))),
        )
        sensor.process(telegram)
        after_update_callback.assert_called_with(sensor)
        assert sensor.last_telegram == telegram
        # every telegram shall trigger callback
        after_update_callback.reset_mock()
        sensor.process(telegram)
        after_update_callback.assert_called_with(sensor)

    #
    # TEST SET
    #
    async def test_set_0(self) -> None:
        """Test set with raw value."""
        xknx = XKNX()
        raw_value = RawValue(xknx, "TestSensor", 0, group_address="1/2/3")
        await raw_value.set(True)
        assert xknx.telegrams.qsize() == 1

        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(True)),
        )

    async def test_set_1(self) -> None:
        """Test set with raw value."""
        xknx = XKNX()
        raw_value = RawValue(xknx, "TestSensor", 1, group_address="1/2/3")
        await raw_value.set(75)
        assert xknx.telegrams.qsize() == 1

        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x4B,))),
        )

    def test_string(self) -> None:
        """Test RawValue string representation."""

        xknx = XKNX()
        value = RawValue(xknx, "Raw", 1, group_address="1/2/3")
        value.remote_value.value = 4
        assert (
            str(value)
            == '<RawValue name="Raw" addresses=<1/2/3, None, [], 4 /> value=4/>'
        )



"""Unit test for RemoteValueString objects."""

import pytest

from xknx import XKNX
from xknx.dpt import DPTArray, DPTBinary, DPTLatin1, DPTString
from xknx.exceptions import ConversionError
from xknx.remote_value import RemoteValueString
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueWrite


class TestRemoteValueString:
    """Test class for RemoteValueString objects."""

    def test_to_knx(self) -> None:
        """Test to_knx function with normal operation."""
        xknx = XKNX()
        dpt_array_string = DPTArray(
            (
                0x4B,
                0x4E,
                0x58,
                0x20,
                0x69,
                0x73,
                0x20,
                0x4F,
                0x4B,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
            )
        )
        remote_value_default = RemoteValueString(xknx)
        assert remote_value_default.dpt_class == DPTString
        assert remote_value_default.to_knx("KNX is OK") == dpt_array_string

        remote_value_ascii = RemoteValueString(xknx, value_type="string")
        assert remote_value_ascii.dpt_class == DPTString
        assert remote_value_ascii.to_knx("KNX is OK") == dpt_array_string

        remote_value_latin1 = RemoteValueString(xknx, value_type="latin_1")
        assert remote_value_latin1.dpt_class == DPTLatin1
        assert remote_value_latin1.to_knx("KNX is OK") == dpt_array_string

    def test_from_knx(self) -> None:
        """Test from_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueString(xknx)
        assert (
            remote_value.from_knx(
                DPTArray(
                    (
                        0x4B,
                        0x4E,
                        0x58,
                        0x20,
                        0x69,
                        0x73,
                        0x20,
                        0x4F,
                        0x4B,
                        0x00,
                        0x00,
                        0x00,
                        0x00,
                        0x00,
                    )
                )
            )
            == "KNX is OK"
        )

    def test_to_knx_error(self) -> None:
        """Test to_knx function with wrong parametern."""
        xknx = XKNX()
        remote_value = RemoteValueString(xknx)
        with pytest.raises(ConversionError):
            remote_value.to_knx("123456789012345")

    async def test_set(self) -> None:
        """Test setting value."""
        xknx = XKNX()
        remote_value = RemoteValueString(xknx, group_address=GroupAddress("1/2/3"))
        remote_value.set("asdf")
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(
                DPTArray((97, 115, 100, 102, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0))
            ),
        )
        remote_value.set("ASDF")
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(
                DPTArray((65, 83, 68, 70, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0))
            ),
        )

    def test_process(self) -> None:
        """Test process telegram."""
        xknx = XKNX()
        remote_value = RemoteValueString(xknx, group_address=GroupAddress("1/2/3"))
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(
                DPTArray(
                    (
                        0x41,
                        0x41,
                        0x41,
                        0x41,
                        0x41,
                        0x42,
                        0x42,
                        0x42,
                        0x42,
                        0x42,
                        0x43,
                        0x43,
                        0x43,
                        0x43,
                    )
                )
            ),
        )
        remote_value.process(telegram)
        assert remote_value.value == "AAAAABBBBBCCCC"

    def test_to_process_error(self) -> None:
        """Test process erroneous telegram."""
        xknx = XKNX()
        remote_value = RemoteValueString(xknx, group_address=GroupAddress("1/2/3"))

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        assert remote_value.process(telegram) is False

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64, 0x65))),
        )
        assert remote_value.process(telegram) is False

        assert remote_value.value is None



"""Implementation of the KNX date data point."""

from __future__ import annotations

from collections.abc import Mapping
from dataclasses import dataclass
import datetime
from typing import Any

from xknx.exceptions import ConversionError

from .dpt import DPTComplex, DPTComplexData
from .payload import DPTArray, DPTBinary


@dataclass(slots=True)
class KNXDate(DPTComplexData):
    """Class for KNX Date."""

    year: int
    month: int
    day: int

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> KNXDate:
        """Init from a dictionary."""
        try:
            year = int(data["year"])
            month = int(data["month"])
            day = int(data["day"])
        except (KeyError, TypeError, ValueError) as err:
            raise ValueError(f"Invalid value for KNXDate: {err}") from err
        return cls(year=year, month=month, day=day)

    def as_dict(self) -> dict[str, int]:
        """Return object as dictionary."""
        return {
            "year": self.year,
            "month": self.month,
            "day": self.day,
        }

    def as_date(self) -> datetime.date:
        """Return datetime object."""
        return datetime.date(self.year, self.month, self.day)

    @classmethod
    def from_date(cls, date: datetime.date) -> KNXDate:
        """Return KNXDate object from a datetime.date object."""
        return cls(date.year, date.month, date.day)


class DPTDate(DPTComplex[KNXDate]):
    """Abstraction for KNX 3 octet date (DPT 11.001)."""

    data_type = KNXDate
    payload_type = DPTArray
    payload_length = 3
    dpt_main_number = 11
    dpt_sub_number = 1
    value_type = "date"

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> KNXDate:
        """Parse/deserialize from KNX/IP raw data."""
        raw = cls.validate_payload(payload)

        day = raw[0] & 0x1F
        month = raw[1] & 0x0F
        year = raw[2] & 0x7F
        if not DPTDate._test_range(day, month, year):
            raise ConversionError("Could not parse DPTDate", raw=raw)

        if year >= 90:
            year += 1900
        else:
            year += 2000

        return KNXDate(year, month, day)

    @classmethod
    def _to_knx(cls, value: KNXDate) -> DPTArray:
        """Serialize to KNX/IP raw data."""

        def _knx_year(year: int) -> int:
            if 2000 <= year < 2090:
                return year - 2000
            if 1990 <= year < 2000:
                return year - 1900
            raise ConversionError(
                f"Could not serialize {cls.dpt_name()}. Year out of range 1990..2089",
                year=year,
            )

        knx_year = _knx_year(value.year)

        if not DPTDate._test_range(value.day, value.month, knx_year):
            raise ConversionError(
                f"Could not serialize {cls.dpt_name()}. Value out of range", value=value
            )

        return DPTArray(
            (
                value.day,
                value.month,
                knx_year,
            )
        )

    @staticmethod
    def _test_range(day: int, month: int, knx_year: int) -> bool:
        """Test if the values are in the correct range."""
        return 1 <= day <= 31 and 1 <= month <= 12 and 0 <= knx_year <= 99



"""Unit test for KNX/IP SearchRequest Extended objects."""

from xknx.knxip import HPAI, KNXIPFrame, SearchRequestExtended
from xknx.knxip.srp import SRP


class TestKNXIPSearchRequestExtended:
    """Test class for KNX/IP SearchRequest Extended objects."""

    def test_search_request_extended(self) -> None:
        """Test parsing and streaming SearchRequest Extended KNX/IP packet."""
        raw = bytes.fromhex("06 10 02 0b 00 0e 08 01 e0 00 17 0C 0E 57")
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, SearchRequestExtended)
        assert knxipframe.body.discovery_endpoint == HPAI(
            ip_addr="224.0.23.12", port=3671
        )

        knxipframe2 = KNXIPFrame.init_from_body(
            SearchRequestExtended(
                discovery_endpoint=HPAI(ip_addr="224.0.23.12", port=3671)
            )
        )

        assert knxipframe2.to_knx() == raw

    def test_search_request_extended_srp(self) -> None:
        """Test parsing and streaming SearchRequest Extended with SRPs KNX/IP packet."""
        raw = bytes.fromhex("06 10 02 0b 00 10 08 01 e0 00 17 0c 0e 57 02 81")
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, SearchRequestExtended)
        assert knxipframe.body.discovery_endpoint == HPAI(
            ip_addr="224.0.23.12", port=3671
        )
        assert len(knxipframe.body.srps) == 1
        assert knxipframe.body.srps[0] == SRP.with_programming_mode()

        knxipframe2 = KNXIPFrame.init_from_body(
            SearchRequestExtended(
                discovery_endpoint=HPAI(ip_addr="224.0.23.12", port=3671),
                srps=[SRP.with_programming_mode()],
            )
        )

        assert knxipframe2.to_knx() == raw

    def test_search_request_extended_multiple_srp(self) -> None:
        """Test parsing and streaming SearchRequest Extended with SRPs KNX/IP packet."""
        raw = bytes.fromhex(
            "06 10 02 0b 00 18 08 01 e0 00 17 0c 0e 57 02 81 08 82 01 02 03 04 05 06"
        )
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, SearchRequestExtended)
        assert knxipframe.body.discovery_endpoint == HPAI(
            ip_addr="224.0.23.12", port=3671
        )
        assert len(knxipframe.body.srps) == 2
        assert knxipframe.body.srps[0] == SRP.with_programming_mode()
        assert knxipframe.body.srps[1] == SRP.with_mac_address(
            bytes([1, 2, 3, 4, 5, 6])
        )

        knxipframe2 = KNXIPFrame.init_from_body(
            SearchRequestExtended(
                discovery_endpoint=HPAI(ip_addr="224.0.23.12", port=3671),
                srps=[
                    SRP.with_programming_mode(),
                    SRP.with_mac_address(bytes([1, 2, 3, 4, 5, 6])),
                ],
            )
        )

        assert knxipframe2.to_knx() == raw



"""Manages global tasks."""

from __future__ import annotations

import asyncio
from collections.abc import Callable, Coroutine, Generator
import logging
from typing import TYPE_CHECKING, Any

from xknx.core import XknxConnectionState

AsyncCallbackType = Callable[[], Coroutine[Any, Any, None]]

if TYPE_CHECKING:
    from xknx import XKNX

logger = logging.getLogger("xknx.log")


class Task:
    """Handles a given task."""

    __slots__ = (
        "_task",
        "name",
        "repeat_after",
        "restart_after_reconnect",
        "target",
        "wait_before_start",
        "wait_for_connection",
        "xknx",
    )

    def __init__(
        self,
        name: str,
        target: AsyncCallbackType | Callable[[], None],
        *,
        restart_after_reconnect: bool = False,
        wait_before_start: float = 0,
        wait_for_connection: bool = False,
        repeat_after: float | None = None,
    ) -> None:
        """
        Initialize a task that can be managed by the TaskRegistry.

        The task execution order is: wait_before_start delay
        → wait_for_connection (if enabled)
        → execute target
        → repeat_after interval (if enabled).

        Args:
            name: Unique identifier for the task, used in logging and task management.
            target: The function to execute. Can be async or sync. Sync functions will run in the main
                event loop thread so shall not perform blocking operations.
            restart_after_reconnect: When True, automatically cancels and restarts the task when the
                KNX bus connection is lost and reestablished. Wait_before_start and wait_for_connection options
                will apply on each restart.
            wait_before_start: Initial delay in seconds before the task begins execution. Applied before
                checking for connection if wait_for_connection is also enabled.
            wait_for_connection: When True, blocks task execution until a KNX bus connection is established.
                The task will wait indefinitely until connected.
            repeat_after: Interval in seconds to repeat the task execution. If None, the task will run only once.

        """
        self.xknx: XKNX | None = None  # used as flag for registration in TaskRegistry
        self.name = name
        self.target = target
        self.restart_after_reconnect = restart_after_reconnect
        self.wait_before_start = wait_before_start
        self.wait_for_connection = wait_for_connection
        self.repeat_after = repeat_after
        self._task: asyncio.Task[None] | None = None

    def _start(self) -> None:
        """Start a task. Shall be called by TaskRegistry.start_task() to ensure proper registration."""
        if self.xknx is None:
            raise RuntimeError("Task must be registered before start().")
        self._task = asyncio.create_task(self._start_internal(), name=self.name)

    async def _start_internal(self) -> None:
        """Start a task and handle options."""
        while True:
            if self.wait_before_start:
                await asyncio.sleep(self.wait_before_start)
            if self.wait_for_connection:
                assert self.xknx is not None
                if not self.xknx.connection_manager.connected.is_set():
                    if self.restart_after_reconnect:
                        # self.reconnected() will restart the task when connection is
                        # reestablished. This will trigger wait_before_start again.
                        return
                    await self.xknx.connection_manager.connected.wait()

            job = self.target()
            if asyncio.iscoroutine(job):
                await job

            if self.repeat_after is None:
                break
            await asyncio.sleep(self.repeat_after)

    def __await__(self) -> Generator[None, None, None]:
        """Wait for task to be finished."""
        if self._task:
            yield from self._task

    def cancel(self) -> None:
        """Cancel a task."""
        if self._task:
            self._task.cancel()
            self._task = None

    def restart(self) -> None:
        """Restart a task."""
        self.cancel()
        self._start()

    def done(self) -> bool:
        """Return if task is finished."""
        return self._task is None or self._task.done()

    def connection_lost(self) -> None:
        """Cancel a task if connection was lost and the task should be cancelled if no connection is established."""
        if self.restart_after_reconnect and self._task:
            logger.debug("Stopping task %s because of connection loss.", self.name)
            self.cancel()

    def reconnected(self) -> None:
        """Restart when reconnected to bus."""
        if self.restart_after_reconnect:
            logger.debug(
                "Restarting task %s as the connection to the bus was reestablished.",
                self.name,
            )
            self.restart()


class TaskRegistry:
    """Manages async tasks in XKNX."""

    __slots__ = ("_background_task", "tasks", "xknx")

    def __init__(self, xknx: XKNX) -> None:
        """Initialize TaskRegistry class."""
        self.xknx = xknx
        self.tasks: set[Task] = set()

        self._background_task: set[asyncio.Task[None]] = set()

    def start_task(self, task: Task) -> None:
        """Register a task and start it. If it is already running, it will be restarted."""
        self.remove_task(task)
        _task = task
        # set xknx to flag registration and allow starting the task
        _task.xknx = self.xknx
        self.tasks.add(_task)
        # ruff: noqa: SLF001
        _task._start()  # pylint: disable=protected-access

    def remove_task(self, task: Task) -> None:
        """Cancel and unregister task."""
        if task in self.tasks:
            task.cancel()
            task.xknx = None  # reset xknx to flag unregistration
            self.tasks.remove(task)

    def start(self) -> None:
        """Start task registry."""
        self.xknx.connection_manager.register_connection_state_changed_cb(
            self.connection_state_changed_cb
        )

    def stop(self) -> None:
        """Stop task registry and cancel all tasks."""
        self.xknx.connection_manager.unregister_connection_state_changed_cb(
            self.connection_state_changed_cb
        )

        for task in self.tasks:
            task.cancel()
        self.tasks = set()

    async def block_till_done(self) -> None:
        """Await all tracked tasks."""
        await asyncio.gather(*self.tasks)

    def connection_state_changed_cb(self, state: XknxConnectionState) -> None:
        """Handle connection state changes."""
        for task in self.tasks:
            if state == XknxConnectionState.CONNECTED:
                task.reconnected()
            else:
                task.connection_lost()

    def background(self, async_func: Coroutine[Any, Any, None]) -> None:
        """
        Run an async task in the background. Task will not be tracked by the TaskRegistry.

        This is a helper method for keeping references to asyncio tasks to prevent
        them from being garbage collected while they are still running.
        """
        # Add task to the set. This creates a strong reference so it can't be garbage collected.
        task = asyncio.create_task(async_func)
        # To prevent keeping references to finished tasks forever,
        self._background_task.add(task)
        # make each task remove its own reference from the set after
        # completion:
        task.add_done_callback(self._background_task.discard)



"""Helper functions for XKNX io module."""

from __future__ import annotations

import asyncio
import ipaddress
import logging
import socket
from typing import cast

import ifaddr

from xknx.exceptions import XKNXException

from .const import DEFAULT_MCAST_GRP, DEFAULT_MCAST_PORT

logger = logging.getLogger("xknx.log")


async def get_default_local_ip(remote_ip: str = DEFAULT_MCAST_GRP) -> str | None:
    """Return the local ip used for communication with remote_ip."""
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
        sock.setblocking(False)  # must be non-blocking for async
        loop = asyncio.get_running_loop()
        try:
            await loop.sock_connect(sock, (remote_ip, DEFAULT_MCAST_PORT))
            local_ip = sock.getsockname()[0]
            logger.debug("Using local ip: %s", local_ip)
            return local_ip  # type: ignore[no-any-return]
        except Exception:  # pylint: disable=broad-except
            logger.warning(
                "The system could not auto detect the source ip for %s on your operating system",
                remote_ip,
            )
            return None


def get_local_ips() -> list[ifaddr.IP]:
    """Return list of local IPv4 addresses."""
    return [ip for iface in ifaddr.get_adapters() for ip in iface.ips if ip.is_IPv4]


def get_local_interface_name(local_ip: str) -> str:
    """Return the name of the interface with the given ip."""
    return next((link.nice_name for link in get_local_ips() if link.ip == local_ip), "")


def get_ip_for_adapter_name(name: str) -> str | None:
    """Return the ip for the given interface name."""
    return next(
        (
            ip.ip
            for iface in ifaddr.get_adapters()
            if name in (iface.name, iface.nice_name)
            for ip in iface.ips
            if isinstance(ip.ip, str)  # IPv6 would return tuple
        ),
        None,
    )


def find_local_ip(gateway_ip: str) -> str | None:
    """Find local IP address on same subnet as gateway."""
    gateway = ipaddress.IPv4Address(gateway_ip)
    for link in get_local_ips():
        network = ipaddress.IPv4Network((link.ip, link.network_prefix), strict=False)
        if gateway in network:
            logger.debug("Using interface: %s", link.nice_name)
            return cast(str, link.ip)
    logger.debug("No interface on same subnet as gateway found.")
    return None


async def validate_ip(address: str, address_name: str = "IP address") -> str:
    """
    Return IPv4 address parsed or resolved as a string.

    Valid addresses are IPv4 strings, adapter names or hostnames.
    Raises XKNXException if address is not a valid IPv4 address or cannot be resolved.
    """
    try:
        ipaddress.IPv4Address(address)
        return address
    except ipaddress.AddressValueError as ex:
        logger.debug(
            "%s is not a valid IPv4 address: %s. Trying to resolve...",
            address_name,
            ex,
        )
    if adapter_ip := get_ip_for_adapter_name(address):
        return adapter_ip
    try:
        return await asyncio.to_thread(socket.gethostbyname, address)
    except socket.gaierror as ex:
        raise XKNXException(f"Could not resolve {address_name}: {address}") from ex



"""Tests for KNX Data Secure."""

import asyncio
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from xknx import XKNX
from xknx.cemi import CEMIFrame, CEMILData, CEMIMessageCode
from xknx.dpt import DPTArray
from xknx.exceptions import DataSecureError
from xknx.secure.data_secure import is_data_secure
from xknx.secure.data_secure_asdu import (
    SecurityAlgorithmIdentifier,
    SecurityALService,
    SecurityControlField,
)
from xknx.secure.keyring import Keyring, sync_load_keyring
from xknx.telegram import (
    GroupAddress,
    IndividualAddress,
    Telegram,
    TelegramDirection,
    apci,
    tpci,
)


@pytest.fixture(name="test_group_response_cemi")
def fixture_test_group_response_cemi() -> CEMIFrame:
    """Return a CEMI frame for a group response telegram."""
    # src = 4.0.9; dst = 0/4/0; GroupValueResponse; value=(116, 41, 41)
    # A+C; seq_num=155806854986
    return CEMIFrame.from_knx(
        bytes.fromhex("29003ce0400904001103f110002446cfef4ac085e7092ab062b44d")
    )


@pytest.fixture(name="test_point_to_point_cemi")
def fixture_test_point_to_point_cemi() -> CEMIFrame:
    """Return a CEMI frame for a group response telegram."""
    # Property Value Write PID_GRP_KEY_TABLE connectionless
    # Object Idx = 5, PropId = 35h, Element Count = 1, Index = 1
    # Data = 20 21 22 23 24 25 26 27 28 29 2A 2B 2C 2D 2E 2F
    # A+C
    # from AN158 v07 KNX Data Security AS - Annex A example
    return CEMIFrame.from_knx(
        bytes.fromhex(
            "29 00 b0 60 ff 67 ff 00 22 03 f1 90 00 00 00 00"
            "00 04 67 67 24 2a 23 08 ca 76 a1 17 74 21 4e e4"
            "cf 5d 94 90 9f 74 3d 05 0d 8f c1 68"
        )
    )


class TestDataSecure:
    """Test class for KNX Data Secure."""

    secure_test_keyring: Keyring

    @classmethod
    def setup_class(cls) -> None:
        """Set up any state specific to the execution of the given class."""
        secure_test_keyfile = Path(__file__).parent / "resources/SecureTest.knxkeys"
        cls.secure_test_keyring = sync_load_keyring(secure_test_keyfile, "test")

    def setup_method(self) -> None:
        """Set up test methods."""
        # pylint: disable=attribute-defined-outside-init
        self.xknx = XKNX()
        self.xknx.knxip_interface = AsyncMock()
        self.xknx.current_address = IndividualAddress("5.0.1")
        self.xknx.cemi_handler.data_secure_init(TestDataSecure.secure_test_keyring)

        self.data_secure = self.xknx.cemi_handler.data_secure

    @patch("xknx.secure.data_secure.DataSecure.outgoing_cemi")
    @patch("xknx.secure.data_secure.DataSecure.received_cemi")
    @patch("xknx.cemi.cemi_handler.CEMIHandler.telegram_received")
    async def test_data_secure_init(
        self,
        mock_telegram_received: MagicMock,  # suppress forwarding to telegrams/management
        mock_ds_received_cemi: MagicMock,
        mock_ds_outgoing_cemi: MagicMock,
    ) -> None:
        """Test DataSecure init and passing frames from CEMIHandler to DataSecure."""
        assert self.data_secure is not None

        assert len(self.data_secure._group_key_table) == 4
        for ga_raw in [1024, 1027, 1028, 1029]:
            assert GroupAddress(ga_raw) in self.data_secure._group_key_table

        assert len(self.data_secure._individual_address_table) == 5
        for ia_raw in ["4.0.0", "4.0.1", "4.0.9", "5.0.0", "5.0.1"]:
            assert (
                IndividualAddress(ia_raw) in self.data_secure._individual_address_table
            )

        # this is based on clock milliseconds
        assert self.data_secure._sequence_number_sending > 0

        # Outgoing
        test_telegram = Telegram(
            destination_address=GroupAddress("0/4/0"),
            payload=apci.GroupValueRead(),
        )

        task = asyncio.create_task(self.xknx.cemi_handler.send_telegram(test_telegram))
        await asyncio.sleep(0)
        # Frame is passed to DataSecure class. Encryption is not tested here.
        mock_ds_outgoing_cemi.assert_called_once()
        assert isinstance(
            mock_ds_outgoing_cemi.call_args.kwargs["cemi_data"], CEMILData
        )
        self.xknx.cemi_handler._l_data_confirmation_event.set()
        await task

        # Incoming
        test_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_IND,
            data=CEMILData.init_from_telegram(test_telegram),
        )
        # Reuse incoming plain APDU. Decryption is not tested here
        mock_ds_received_cemi.return_value = test_cemi.data

        self.xknx.cemi_handler.handle_cemi_frame(test_cemi)
        mock_ds_received_cemi.assert_called_once()
        assert isinstance(
            mock_ds_received_cemi.call_args.kwargs["cemi_data"], CEMILData
        )
        mock_telegram_received.assert_called_once()

    def test_data_secure_init_invalid_system_time(self) -> None:
        """Test DataSecure init with invalid system time."""
        with (
            patch("time.time", return_value=1515108203.0),  # 2018-01-04T23:23:23+00:00
            pytest.raises(DataSecureError, match=r"Initial sequence number out of .*"),
        ):
            self.xknx.cemi_handler.data_secure_init(TestDataSecure.secure_test_keyring)

    def test_data_secure_group_send(self) -> None:
        """Test outgoing DataSecure group communication."""
        self.data_secure._sequence_number_sending = 160170101607

        test_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_REQ,
            data=CEMILData.init_from_telegram(
                Telegram(
                    destination_address=GroupAddress("0/4/0"),
                    payload=apci.GroupValueRead(),
                ),
                src_addr=self.xknx.current_address,
            ),
        )
        secured_frame_data = self.data_secure.outgoing_cemi(test_cemi.data)
        assert isinstance(secured_frame_data, CEMILData)
        assert isinstance(secured_frame_data.payload, apci.SecureAPDU)
        secured_asdu = secured_frame_data.payload.secured_data

        assert int.from_bytes(secured_asdu.sequence_number_bytes, "big") == 160170101607
        assert secured_asdu.secured_apdu == bytes.fromhex("cd18")
        assert secured_asdu.message_authentication_code == bytes.fromhex("4afe5744")
        # sequence number sending was incremented
        assert self.data_secure._sequence_number_sending == 160170101608

        assert secured_frame_data.to_knx() == bytes.fromhex(
            "bce0500104000e03f11000254ae1cb67cd184afe5744"
        )

    def test_data_secure_group_receive(
        self, test_group_response_cemi: CEMIFrame
    ) -> None:
        """Test incoming DataSecure group communication."""
        assert (
            self.data_secure._individual_address_table[IndividualAddress("4.0.9")]
            == 155806854915
        )
        assert isinstance(test_group_response_cemi.data, CEMILData)
        assert is_data_secure(test_group_response_cemi.data)
        assert test_group_response_cemi.data.src_addr == IndividualAddress("4.0.9")
        assert isinstance(test_group_response_cemi.data.payload, apci.SecureAPDU)

        plain_frame_data = self.data_secure.received_cemi(test_group_response_cemi.data)
        assert isinstance(plain_frame_data, CEMILData)
        assert not is_data_secure(plain_frame_data)
        assert plain_frame_data.payload == apci.GroupValueResponse(
            DPTArray((116, 41, 41))
        )
        # individual_address_table sequence number was updated
        assert (
            self.data_secure._individual_address_table[IndividualAddress("4.0.9")]
            == 155806854986
        )

    def test_data_secure_individual_receive_tool_key(
        self, test_point_to_point_cemi: CEMIFrame
    ) -> None:
        """Test incoming DataSecure point-to-point communication via tool key."""
        self.xknx.current_address = IndividualAddress("15.15.0")
        assert isinstance(test_point_to_point_cemi.data.payload, apci.SecureAPDU)

        with pytest.raises(
            DataSecureError, match=r"System broadcast and tool access not supported.*"
        ):
            self.data_secure.received_cemi(test_point_to_point_cemi.data)
        # don't raise through handle_cemi_frame()
        assert (
            self.xknx.cemi_handler.handle_cemi_frame(test_point_to_point_cemi) is None
        )

    def test_data_secure_individual_receive(
        self, test_point_to_point_cemi: CEMIFrame
    ) -> None:
        """Test incoming DataSecure point-to-point communication."""
        self.xknx.current_address = IndividualAddress("15.15.0")
        assert isinstance(test_point_to_point_cemi.data.payload, apci.SecureAPDU)
        # don't use tool key or system broadcast
        # further validation is skipped so we can use the same test data
        test_point_to_point_cemi.data.payload.scf.tool_access = False
        test_point_to_point_cemi.data.payload.scf.system_broadcast = False
        with pytest.raises(
            DataSecureError,
            match=r"Secure Point-to-Point communication not supported.*",
        ):
            self.data_secure.received_cemi(test_point_to_point_cemi.data)
        # don't raise through handle_cemi_frame()
        assert (
            self.xknx.cemi_handler.handle_cemi_frame(test_point_to_point_cemi) is None
        )

    def test_data_secure_group_receive_unknown_source(
        self, test_group_response_cemi: CEMIFrame
    ) -> None:
        """Test incoming DataSecure group communication from unknown source."""
        test_group_response_cemi.data.src_addr = IndividualAddress("1.2.3")
        with pytest.raises(
            DataSecureError,
            match=r"Source address not found in Security Individual Address Table.*",
        ):
            self.data_secure.received_cemi(test_group_response_cemi.data)

    def test_data_secure_group_receive_unknown_destination(
        self, test_group_response_cemi: CEMIFrame
    ) -> None:
        """Test incoming DataSecure group communication for unknown destination."""
        test_group_response_cemi.data.dst_addr = GroupAddress("1/2/3")
        with pytest.raises(
            DataSecureError,
            match=r"No key found for group address.*",
        ):
            self.data_secure.received_cemi(test_group_response_cemi.data)

    def test_data_secure_group_receive_wrong_sequence_number(
        self, test_group_response_cemi: CEMIFrame
    ) -> None:
        """Test incoming DataSecure group communication with wrong sequence number."""
        seq_num = 155806854986
        assert (
            test_group_response_cemi.data.payload.secured_data.sequence_number_bytes
            == seq_num.to_bytes(6, "big")
        )
        # sequence number already used
        self.data_secure._individual_address_table[IndividualAddress("4.0.9")] = seq_num
        with pytest.raises(
            DataSecureError,
            match=r"Sequence number too low.*",
        ):
            self.data_secure.received_cemi(test_group_response_cemi.data)

    def test_data_secure_group_receive_wrong_mac(
        self, test_group_response_cemi: CEMIFrame
    ) -> None:
        """Test incoming DataSecure group communication with wrong MAC."""
        test_group_response_cemi.data.payload.secured_data.message_authentication_code = bytes(
            4
        )
        with pytest.raises(
            DataSecureError,
            match=r"Data Secure MAC verification failed.*",
        ):
            self.data_secure.received_cemi(test_group_response_cemi.data)

    def test_data_secure_group_receive_plain_frame(self) -> None:
        """Test incoming DataSecure group communication with plain frame."""
        src_addr = IndividualAddress("4.0.9")
        test_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_IND,
            data=CEMILData.init_from_telegram(
                Telegram(
                    destination_address=GroupAddress("0/4/0"),
                    direction=TelegramDirection.INCOMING,
                    payload=apci.GroupValueResponse(DPTArray((116, 41, 41))),
                ),
                src_addr=src_addr,
            ),
        )
        assert src_addr in self.data_secure._individual_address_table
        with pytest.raises(
            DataSecureError,
            match=r"Discarding frame with plain APDU for secure group address.*",
        ):
            self.data_secure.received_cemi(test_cemi.data)

    def test_non_secure_group_receive_plain_frame(self) -> None:
        """Test incoming non-secure group communication with plain frame."""
        dst_addr = GroupAddress("1/2/3")
        test_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_IND,
            data=CEMILData.init_from_telegram(
                Telegram(
                    destination_address=dst_addr,
                    direction=TelegramDirection.INCOMING,
                    payload=apci.GroupValueResponse(DPTArray((116, 41, 41))),
                ),
                src_addr=IndividualAddress("4.0.9"),
            ),
        )
        assert dst_addr not in self.data_secure._group_key_table
        assert self.data_secure.received_cemi(test_cemi.data) == test_cemi.data

    def test_non_secure_group_send_plain_frame(self) -> None:
        """Test outgoing non-secure group communication with plain frame."""
        dst_addr = GroupAddress("1/2/3")
        test_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_REQ,
            data=CEMILData.init_from_telegram(
                Telegram(
                    destination_address=dst_addr,
                    direction=TelegramDirection.OUTGOING,
                    payload=apci.GroupValueResponse(DPTArray((116, 41, 41))),
                ),
                src_addr=self.xknx.current_address,
            ),
        )
        assert dst_addr not in self.data_secure._group_key_table
        assert self.data_secure.outgoing_cemi(test_cemi.data) == test_cemi.data

    def test_non_secure_individual_receive_plain_frame(self) -> None:
        """Test incoming non-secure group communication with plain frame."""
        src_addr = IndividualAddress("1.2.3")
        test_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_IND,
            data=CEMILData.init_from_telegram(
                Telegram(
                    destination_address=self.xknx.current_address,
                    direction=TelegramDirection.INCOMING,
                    payload=apci.PropertyValueWrite(
                        object_index=5,
                        property_id=0x35,
                        count=1,
                        start_index=1,
                        data=bytes.fromhex(
                            "20 21 22 23 24 25 26 27 28 29 2A 2B 2C 2D 2E 2F"
                        ),
                    ),
                    tpci=tpci.TDataIndividual(),
                ),
                src_addr=IndividualAddress("4.0.9"),
            ),
        )
        assert src_addr not in self.data_secure._individual_address_table
        assert self.data_secure.received_cemi(test_cemi.data) == test_cemi.data

    def test_non_secure_individual_send_plain_frame(self) -> None:
        """Test outgoing non-secure group communication with plain frame."""
        dst_addr = IndividualAddress("1.2.3")
        test_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_REQ,
            data=CEMILData.init_from_telegram(
                Telegram(
                    destination_address=dst_addr,
                    direction=TelegramDirection.INCOMING,
                    payload=apci.PropertyValueWrite(
                        object_index=5,
                        property_id=0x35,
                        count=1,
                        start_index=1,
                        data=bytes.fromhex(
                            "20 21 22 23 24 25 26 27 28 29 2A 2B 2C 2D 2E 2F"
                        ),
                    ),
                    tpci=tpci.TDataIndividual(),
                ),
                src_addr=self.xknx.current_address,
            ),
        )
        assert dst_addr not in self.data_secure._individual_address_table
        assert self.data_secure.outgoing_cemi(test_cemi.data) == test_cemi.data

    def test_data_secure_authentication_only(self) -> None:
        """Test frame de-/serialization for DataSecure authentication only."""
        # This is currently not used from xknx and I also don't know if it is used
        # in any ETS or runtime KNX communication. Therefore a very generic test.
        dst_addr = GroupAddress("0/4/0")
        test_telegram = Telegram(
            destination_address=dst_addr,
            direction=TelegramDirection.OUTGOING,
            payload=apci.GroupValueWrite(DPTArray((1, 2))),
        )
        test_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_REQ,
            data=CEMILData.init_from_telegram(
                telegram=test_telegram,
                src_addr=self.xknx.current_address,
            ),
        )
        scf = SecurityControlField(
            algorithm=SecurityAlgorithmIdentifier.CCM_AUTHENTICATION,
            service=SecurityALService.S_A_DATA,
            system_broadcast=False,
            tool_access=False,
        )
        key = self.data_secure._group_key_table[dst_addr]

        outgoing_signed_cemi_data = self.data_secure._secure_data_cemi(
            key=key, scf=scf, cemi_data=test_cemi.data
        )
        assert outgoing_signed_cemi_data.payload.secured_data is not None

        # create new cemi to avoid mixed bytearray / byte parts
        incoming_cemi = CEMIFrame.from_knx(
            b"\x11\x00" + outgoing_signed_cemi_data.to_knx()
        )
        assert isinstance(incoming_cemi.data, CEMILData)
        # receive same cemi - fake individual address table entry
        self.data_secure._individual_address_table[incoming_cemi.data.src_addr] = 1
        assert self.data_secure.received_cemi(incoming_cemi.data) == test_cemi.data

        # Test wrong MAC
        self.data_secure._individual_address_table[incoming_cemi.data.src_addr] = 1
        incoming_cemi.data.payload.secured_data.message_authentication_code = bytes(4)
        with pytest.raises(
            DataSecureError,
            match=r"Data Secure MAC verification failed.*",
        ):
            self.data_secure.received_cemi(incoming_cemi.data)



"""Unit test for KNX/IP DeviceConfigurationRequest objects."""

import pytest

from xknx.cemi import CEMIFrame, CEMIMessageCode, CEMIMPropInfo, CEMIMPropReadRequest
from xknx.exceptions import CouldNotParseKNXIP
from xknx.knxip import DeviceConfigurationRequest, KNXIPFrame
from xknx.profile.const import ResourceKNXNETIPPropertyId, ResourceObjectType


class TestKNXIPDeviceConfigurationRequest:
    """Test class for KNX/IP DeviceConfigurationRequest objects."""

    def test_device_configuration_request(self) -> None:
        """Test parsing and streaming device configuration ACK KNX/IP packet."""
        raw = bytes.fromhex("06 10 03 10 00 11 04 2A 17 00 FC 00 0B 01 34 10 01")
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, DeviceConfigurationRequest)
        assert knxipframe.body.communication_channel_id == 42
        assert knxipframe.body.sequence_counter == 23
        assert isinstance(knxipframe.body.raw_cemi, bytes)

        incoming_cemi = CEMIFrame.from_knx(knxipframe.body.raw_cemi)
        assert incoming_cemi.code == CEMIMessageCode.M_PROP_READ_REQ
        assert incoming_cemi.data == CEMIMPropReadRequest(
            property_info=CEMIMPropInfo(
                object_type=ResourceObjectType.OBJECT_KNXNETIP_PARAMETER,
                object_instance=1,
                property_id=ResourceKNXNETIPPropertyId.PID_KNX_INDIVIDUAL_ADDRESS,
                start_index=1,
                number_of_elements=1,
            )
        )

        outgoing_cemi = CEMIFrame(
            code=CEMIMessageCode.M_PROP_READ_REQ,
            data=CEMIMPropReadRequest(
                property_info=CEMIMPropInfo(
                    object_type=ResourceObjectType.OBJECT_KNXNETIP_PARAMETER,
                    object_instance=1,
                    property_id=ResourceKNXNETIPPropertyId.PID_KNX_INDIVIDUAL_ADDRESS,
                    start_index=1,
                    number_of_elements=1,
                )
            ),
        )
        device_configuration_req = DeviceConfigurationRequest(
            communication_channel_id=42,
            sequence_counter=23,
            raw_cemi=outgoing_cemi.to_knx(),
        )
        knxipframe2 = KNXIPFrame.init_from_body(device_configuration_req)

        assert knxipframe2.to_knx() == raw

    def test_from_knx_wrong_header(self) -> None:
        """Test parsing and streaming wrong DeviceConfigurationRequest (wrong length byte)."""
        raw = bytes((0x06, 0x10, 0x03, 0x10, 0x00, 0x15, 0x03))
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)

    def test_from_knx_wrong_header2(self) -> None:
        """Test parsing and streaming wrong DeviceConfigurationRequest (wrong length)."""
        raw = bytes((0x06, 0x10, 0x03, 0x10, 0x00, 0x15, 0x04))
        with pytest.raises(CouldNotParseKNXIP):
            KNXIPFrame.from_knx(raw)



"""Unit test for KNX/IP Tunnelling Request/Response."""

from unittest.mock import patch

from xknx.cemi import CEMIFrame, CEMILData, CEMIMessageCode
from xknx.dpt import DPTArray
from xknx.io.request_response import Tunnelling
from xknx.io.transport import UDPTransport
from xknx.knxip import (
    HPAI,
    ConnectionStateRequest,
    ErrorCode,
    KNXIPFrame,
    TunnellingAck,
    TunnellingRequest,
)
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueWrite


class TestTunnelling:
    """Test class for xknx/io/Tunnelling objects."""

    async def test_tunnelling(self) -> None:
        """Test tunnelling from KNX bus."""
        data_endpoint = ("192.168.1.2", 4567)
        udp_transport = UDPTransport(("192.168.1.1", 0), ("192.168.1.2", 1234))
        raw_cemi = CEMIFrame(
            code=CEMIMessageCode.L_DATA_IND,
            data=CEMILData.init_from_telegram(
                Telegram(
                    destination_address=GroupAddress("1/2/3"),
                    payload=GroupValueWrite(DPTArray((0x1, 0x2, 0x3))),
                )
            ),
        ).to_knx()
        tunnelling_request = TunnellingRequest(
            communication_channel_id=23,
            sequence_counter=42,
            raw_cemi=raw_cemi,
        )
        tunnelling = Tunnelling(udp_transport, data_endpoint, tunnelling_request)
        tunnelling.timeout_in_seconds = 0

        assert tunnelling.awaited_response_class == TunnellingAck

        exp_knxipframe = KNXIPFrame.init_from_body(tunnelling_request)
        with (
            patch("xknx.io.transport.UDPTransport.send") as mock_udp_send,
            patch("xknx.io.transport.UDPTransport.getsockname") as mock_udp_getsockname,
        ):
            mock_udp_getsockname.return_value = ("192.168.1.3", 4321)
            await tunnelling.start()
            mock_udp_send.assert_called_with(exp_knxipframe, addr=data_endpoint)

        # Response KNX/IP-Frame with wrong type
        wrong_knxipframe = KNXIPFrame.init_from_body(ConnectionStateRequest())
        with patch("logging.Logger.warning") as mock_warning:
            tunnelling.response_rec_callback(wrong_knxipframe, HPAI(), None)
            mock_warning.assert_called_with("Could not understand knxipframe")

        # Response KNX/IP-Frame with error:
        err_knxipframe = KNXIPFrame.init_from_body(
            TunnellingAck(status_code=ErrorCode.E_CONNECTION_ID)
        )
        with patch("logging.Logger.debug") as mock_warning:
            tunnelling.response_rec_callback(err_knxipframe, HPAI(), None)
            mock_warning.assert_called_with(
                "Error: KNX bus responded to request of type '%s' with error in '%s': %s",
                type(tunnelling).__name__,
                type(err_knxipframe.body).__name__,
                ErrorCode.E_CONNECTION_ID,
            )

        # Correct Response KNX/IP-Frame:
        res_knxipframe = KNXIPFrame.init_from_body(TunnellingAck())
        tunnelling.response_rec_callback(res_knxipframe, HPAI(), None)
        assert tunnelling.success



"""Unit test for RemoteValueSwitch objects."""

import pytest

from xknx import XKNX
from xknx.dpt import DPTArray, DPTBinary
from xknx.exceptions import ConversionError
from xknx.remote_value import RemoteValueSwitch
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueWrite


class TestRemoteValueSwitch:
    """Test class for RemoteValueSwitch objects."""

    def test_to_knx(self) -> None:
        """Test to_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueSwitch(xknx)
        assert remote_value.to_knx(True) == DPTBinary(True)
        assert remote_value.to_knx(False) == DPTBinary(False)

    def test_from_knx(self) -> None:
        """Test from_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueSwitch(xknx)
        assert remote_value.from_knx(DPTBinary(True)) is True
        assert remote_value.from_knx(DPTBinary(0)) is False

    def test_to_knx_invert(self) -> None:
        """Test to_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueSwitch(xknx, invert=True)
        assert remote_value.to_knx(True) == DPTBinary(0)
        assert remote_value.to_knx(False) == DPTBinary(1)

    def test_from_knx_invert(self) -> None:
        """Test from_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueSwitch(xknx, invert=True)
        assert remote_value.from_knx(DPTBinary(1)) is False
        assert remote_value.from_knx(DPTBinary(0)) is True

    def test_to_knx_error(self) -> None:
        """Test to_knx function with wrong parametern."""
        xknx = XKNX()
        remote_value = RemoteValueSwitch(xknx)
        with pytest.raises(ConversionError):
            remote_value.to_knx(1)

    async def test_set(self) -> None:
        """Test setting value."""
        xknx = XKNX()
        remote_value = RemoteValueSwitch(xknx, group_address=GroupAddress("1/2/3"))
        remote_value.on()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        remote_value.off()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(0)),
        )

    def test_process(self) -> None:
        """Test process telegram."""
        xknx = XKNX()
        remote_value = RemoteValueSwitch(xknx, group_address=GroupAddress("1/2/3"))
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        assert remote_value.value is None
        remote_value.process(telegram)
        assert remote_value.telegram is not None
        assert remote_value.value is True

    def test_process_off(self) -> None:
        """Test process OFF telegram."""
        xknx = XKNX()
        remote_value = RemoteValueSwitch(xknx, group_address=GroupAddress("1/2/3"))
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(0)),
        )
        assert remote_value.value is None
        remote_value.process(telegram)
        assert remote_value.telegram is not None
        assert remote_value.value is False

    def test_to_process_error(self) -> None:
        """Test process erroneous telegram."""
        xknx = XKNX()
        remote_value = RemoteValueSwitch(xknx, group_address=GroupAddress("1/2/3"))

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray(0x01)),
        )
        assert remote_value.process(telegram) is False

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(3)),
        )
        assert remote_value.process(telegram) is False

        assert remote_value.value is None



"""Test for CEMIHandler."""

import asyncio
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from xknx import XKNX
from xknx.cemi import CEMIFrame, CEMILData, CEMIMessageCode
from xknx.dpt import DPTArray
from xknx.exceptions import ConfirmationError
from xknx.telegram import GroupAddress, IndividualAddress, Telegram, apci, tpci

from ..conftest import EventLoopClockAdvancer


async def test_wait_for_l2_confirmation(time_travel: EventLoopClockAdvancer) -> None:
    """Test waiting for L_DATA.con before sending another L_DATA.req."""
    xknx = XKNX()
    xknx.knxip_interface = AsyncMock()

    test_telegram = Telegram(
        destination_address=GroupAddress(1),
        payload=apci.GroupValueWrite(DPTArray((1,))),
    )
    test_cemi = CEMIFrame(
        code=CEMIMessageCode.L_DATA_REQ,
        data=CEMILData.init_from_telegram(test_telegram),
    )
    test_cemi_confirmation = CEMIFrame(
        code=CEMIMessageCode.L_DATA_CON,
        data=CEMILData.init_from_telegram(
            test_telegram,
        ),
    )
    task = asyncio.create_task(xknx.cemi_handler.send_telegram(test_telegram))
    await time_travel(0)
    xknx.knxip_interface.send_cemi.assert_called_once_with(test_cemi)
    assert xknx.connection_manager.cemi_count_outgoing == 0

    assert not task.done()
    xknx.cemi_handler.handle_cemi_frame(test_cemi_confirmation)
    await time_travel(0)
    assert task.done()
    await task
    assert xknx.connection_manager.cemi_count_outgoing == 1
    assert xknx.connection_manager.cemi_count_outgoing_error == 0
    assert test_telegram.data_secure is False

    # no L_DATA.con received -> raise ConfirmationError
    xknx.knxip_interface.send_cemi.reset_mock()
    task = asyncio.create_task(xknx.cemi_handler.send_telegram(test_telegram))
    await time_travel(0)
    xknx.knxip_interface.send_cemi.assert_called_once_with(test_cemi)
    with pytest.raises(ConfirmationError):
        await time_travel(3)
        assert task.done()
        await task
        assert xknx.connection_manager.cemi_count_outgoing == 1
        assert xknx.connection_manager.cemi_count_outgoing_error == 1


@patch("xknx.management.management.Management.process")
def test_incoming_cemi(mock_management_process: MagicMock) -> None:
    """Test incoming CEMI."""
    xknx = XKNX()
    xknx.current_address = IndividualAddress("1.1.1")

    # TDataGroup Telegram
    test_telegram = Telegram(
        destination_address=GroupAddress(1),
        payload=apci.GroupValueWrite(DPTArray((1,))),
    )
    test_group_cemi = CEMIFrame(
        code=CEMIMessageCode.L_DATA_IND,
        data=CEMILData.init_from_telegram(test_telegram),
    )
    xknx.cemi_handler.handle_cemi_frame(test_group_cemi)
    assert xknx.telegrams.qsize() == 1
    mock_management_process.assert_not_called()
    xknx.telegrams.get_nowait()  # remove telegram from queue

    # L_DATA_CON and L_DATA_REQ should not be forwarded to the telegram queue or management
    test_incoming_l_data_con = CEMIFrame(
        code=CEMIMessageCode.L_DATA_CON,
        data=CEMILData.init_from_telegram(test_telegram),
    )
    xknx.cemi_handler.handle_cemi_frame(test_incoming_l_data_con)
    assert not xknx.telegrams.qsize()
    mock_management_process.assert_not_called()
    test_incoming_l_data_req = CEMIFrame(
        code=CEMIMessageCode.L_DATA_REQ,
        data=CEMILData.init_from_telegram(test_telegram),
    )
    xknx.cemi_handler.handle_cemi_frame(test_incoming_l_data_req)
    assert not xknx.telegrams.qsize()
    mock_management_process.assert_not_called()
    assert xknx.connection_manager.cemi_count_incoming == 1


@pytest.mark.parametrize(
    "telegram",
    [
        Telegram(
            destination_address=GroupAddress(0),
            tpci=tpci.TDataBroadcast(),
        ),
        Telegram(
            destination_address=IndividualAddress("1.1.1"),
            tpci=tpci.TConnect(),
        ),
        Telegram(
            destination_address=IndividualAddress("1.1.1"),
            tpci=tpci.TDataIndividual(),
        ),
    ],
)
@patch("xknx.management.management.Management.process")
def test_incoming_management_telegram(
    mock_management_process: MagicMock, telegram: Telegram
) -> None:
    """Test incoming management CEMI."""
    xknx = XKNX()
    xknx.current_address = IndividualAddress("1.1.1")

    test_cemi = CEMIFrame(
        code=CEMIMessageCode.L_DATA_IND,
        data=CEMILData.init_from_telegram(telegram),
    )
    xknx.cemi_handler.handle_cemi_frame(test_cemi)
    mock_management_process.assert_called_once()
    assert xknx.telegrams.qsize() == 0
    assert xknx.connection_manager.cemi_count_incoming == 1


@pytest.mark.parametrize(
    "raw",
    [
        # <CouldNotParseCEMI description="CEMI too small. Length: 9; CEMI: 2900b06010fa10ff00" />
        # communication_channel_id: 0x02   sequence_counter: 0x81
        bytes.fromhex("2900b06010fa10ff00"),
    ],
)
@patch("xknx.cemi.cemi_handler.CEMIHandler.handle_cemi_frame")
@patch("logging.Logger.warning")
def test_invalid_cemi(
    mock_warning: MagicMock, mock_handle_cemi_frame: MagicMock, raw: bytes
) -> None:
    """Test incoming invalid CEMI Frames."""
    xknx = XKNX()

    xknx.cemi_handler.handle_raw_cemi(raw)
    mock_warning.assert_called_once()
    mock_handle_cemi_frame.assert_not_called()
    assert xknx.connection_manager.cemi_count_incoming_error == 1


@pytest.mark.parametrize(
    "raw",
    [
        # LDataInd Unsupported Extended APCI from 0.0.1 to 0/0/0 broadcast
        # <UnsupportedCEMIMessage description="APCI not supported: 0b1111111000 in CEMI: 2900b0d0000100000103f8" />
        bytes.fromhex("2900b0d0000100000103f8"),
    ],
)
@patch("xknx.cemi.cemi_handler.CEMIHandler.handle_cemi_frame")
@patch("logging.Logger.info")
def test_unsupported_cemi(
    mock_info: MagicMock, mock_handle_cemi_frame: MagicMock, raw: bytes
) -> None:
    """Test incoming unsupported CEMI Frames."""
    xknx = XKNX()

    xknx.cemi_handler.handle_raw_cemi(raw)
    mock_info.assert_called_once()
    mock_handle_cemi_frame.assert_not_called()
    assert xknx.connection_manager.cemi_count_incoming_error == 1


@patch("xknx.cemi.cemi_handler.CEMIHandler.telegram_received")
@patch("logging.Logger.debug")
def test_incoming_from_own_ia(
    mock_debug: MagicMock, mock_telegram_received: MagicMock
) -> None:
    """Test incoming CEMI from own IA."""
    xknx = XKNX()
    xknx.current_address = IndividualAddress("1.1.22")
    # L_Data.ind GroupValueWrite from 1.1.22 to to 5/1/22 with DPT9 payload 0C 3F
    raw = bytes.fromhex("2900bcd011162916030080 0c 3f")

    xknx.cemi_handler.handle_raw_cemi(raw)
    mock_debug.assert_called_once()
    mock_telegram_received.assert_called_once()
    assert xknx.connection_manager.cemi_count_incoming == 1
    assert xknx.connection_manager.cemi_count_incoming_error == 0


@pytest.mark.parametrize(
    "raw_cemi_data_secure",
    [
        # src = 4.0.9; dst = 0/4/0; GroupValueResponse; value=(116, 41, 41)
        # A+C; seq_num=155806854986
        bytes.fromhex("29003ce0400904001103f110002446cfef4ac085e7092ab062b44d"),
        # Property Value Write PID_GRP_KEY_TABLE connectionless
        # Object Idx = 5, PropId = 35h, Element Count = 1, Index = 1
        # Data = 20 21 22 23 24 25 26 27 28 29 2A 2B 2C 2D 2E 2F
        # A+C
        # from AN158 v07 KNX Data Security AS - Annex A example
        bytes.fromhex(
            "29 00 b0 60 ff 67 ff 00 22 03 f1 90 00 00 00 00"
            "00 04 67 67 24 2a 23 08 ca 76 a1 17 74 21 4e e4"
            "cf 5d 94 90 9f 74 3d 05 0d 8f c1 68"
        ),
    ],
)
@patch("xknx.cemi.cemi_handler.CEMIHandler.telegram_received")
def test_incoming_cemi_no_data_secure_keys(
    mock_telegram_received: MagicMock,
    raw_cemi_data_secure: bytes,
) -> None:
    """Test incoming DataSecure CEMI when no DataSecure keys are initialized."""
    xknx = XKNX()
    xknx.current_address = IndividualAddress("5.0.1")

    with (
        patch("logging.Logger.debug") as mock_debug,
    ):
        xknx.cemi_handler.handle_raw_cemi(raw_cemi_data_secure)
    assert mock_debug.call_count == 2  # one for reception, one for DataSecure debug
    # frame is dropped because no keys are available
    mock_telegram_received.assert_not_called()
    # not having a key isn't an error per se, so no incoming_error count
    assert xknx.connection_manager.cemi_count_incoming == 1
    assert xknx.connection_manager.cemi_count_incoming_error == 0
    assert xknx.connection_manager.undecoded_data_secure == 1



"""Unit test for KNX 2 byte objects."""

import pytest

from xknx.dpt import DPTArray, DPTUElCurrentmA
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestDPT2byte:
    """Test class for KNX 2 byte objects."""

    #
    # DPTUElCurrentmA
    #
    def test_current_settings(self) -> None:
        """Test members of DPTUElCurrentmA."""
        assert DPTUElCurrentmA.value_min == 0
        assert DPTUElCurrentmA.value_max == 65535
        assert DPTUElCurrentmA.unit == "mA"
        assert DPTUElCurrentmA.resolution == 1

    def test_current_assert_min_exceeded(self) -> None:
        """Test initialization of DPTUElCurrentmA with wrong value (Underflow)."""
        with pytest.raises(ConversionError):
            DPTUElCurrentmA.to_knx(-1)

    def test_current_to_knx_exceed_limits(self) -> None:
        """Test initialization of DPTUElCurrentmA with wrong value (Overflow)."""
        with pytest.raises(ConversionError):
            DPTUElCurrentmA.to_knx(65536)

    def test_current_value_max_value(self) -> None:
        """Test DPTUElCurrentmA parsing and streaming."""
        assert DPTUElCurrentmA.to_knx(65535) == DPTArray((0xFF, 0xFF))
        assert DPTUElCurrentmA.from_knx(DPTArray((0xFF, 0xFF))) == 65535

    def test_current_value_min_value(self) -> None:
        """Test DPTUElCurrentmA parsing and streaming with null values."""
        assert DPTUElCurrentmA.to_knx(0) == DPTArray((0x00, 0x00))
        assert DPTUElCurrentmA.from_knx(DPTArray((0x00, 0x00))) == 0

    def test_current_value_38(self) -> None:
        """Test DPTUElCurrentmA parsing and streaming 38mA."""
        assert DPTUElCurrentmA.to_knx(38) == DPTArray((0x00, 0x26))
        assert DPTUElCurrentmA.from_knx(DPTArray((0x00, 0x26))) == 38

    def test_current_value_78(self) -> None:
        """Test DPTUElCurrentmA parsing and streaming 78mA."""
        assert DPTUElCurrentmA.to_knx(78) == DPTArray((0x00, 0x4E))
        assert DPTUElCurrentmA.from_knx(DPTArray((0x00, 0x4E))) == 78

    def test_current_value_1234(self) -> None:
        """Test DPTUElCurrentmA parsing and streaming 4660mA."""
        assert DPTUElCurrentmA.to_knx(4660) == DPTArray((0x12, 0x34))
        assert DPTUElCurrentmA.from_knx(DPTArray((0x12, 0x34))) == 4660

    def test_current_wrong_value_from_knx(self) -> None:
        """Test DPTUElCurrentmA parsing with wrong value."""
        with pytest.raises(CouldNotParseTelegram):
            DPTUElCurrentmA.from_knx(DPTArray((0xFF, 0x4E, 0x12)))



"""
Module for managing a raw value via KNX.

It provides functionality for

* reading the current value from KNX bus or providing a value to the bus.
* send local state changes to KNX bus.
* watching for state updates from KNX bus.
"""

from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING

from xknx.remote_value import GroupAddressesType, RemoteValueRaw

from .device import Device, DeviceCallbackType

if TYPE_CHECKING:
    from xknx.telegram import Telegram
    from xknx.xknx import XKNX


class RawValue(Device):
    """Class for managing a raw value."""

    def __init__(
        self,
        xknx: XKNX,
        name: str,
        payload_length: int,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        respond_to_read: bool = False,
        sync_state: bool | int | float | str = True,
        always_callback: bool = False,
        device_updated_cb: DeviceCallbackType[RawValue] | None = None,
    ) -> None:
        """Initialize Sensor class."""
        super().__init__(xknx, name, device_updated_cb)
        self.always_callback = always_callback
        self.respond_to_read = respond_to_read
        self.remote_value = RemoteValueRaw(
            xknx,
            payload_length=payload_length,
            group_address=group_address,
            group_address_state=group_address_state,
            sync_state=sync_state,
            device_name=self.name,
            after_update_cb=self.after_update,
        )

    def _iter_remote_values(self) -> Iterator[RemoteValueRaw]:
        """Iterate the devices RemoteValue classes."""
        yield self.remote_value

    @property
    def last_telegram(self) -> Telegram | None:
        """Return the last telegram received from the RemoteValue."""
        return self.remote_value.telegram

    def process_group_write(self, telegram: Telegram) -> None:
        """Process incoming and outgoing GROUP WRITE telegram."""
        self.remote_value.process(telegram, always_callback=self.always_callback)

    def process_group_read(self, telegram: Telegram) -> None:
        """Process incoming GroupValueResponse telegrams."""
        if (
            self.respond_to_read
            and telegram.destination_address == self.remote_value.group_address
        ):
            self.remote_value.respond()

    async def set(self, value: int) -> None:
        """Set new value."""
        self.remote_value.set(value)

    def resolve_state(self) -> int | None:
        """Return the current state of the sensor as an unsigned integer."""
        return self.remote_value.value

    def __str__(self) -> str:
        """Return object as readable string."""
        return (
            f'<RawValue name="{self.name}" '
            f"addresses={self.remote_value.group_addr_str()} "
            f"value={self.resolve_state().__repr__()}/>"
        )



"""Cryptographical calculations of KNX specification example frames."""

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric.x25519 import (
    X25519PrivateKey,
    X25519PublicKey,
)
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


def bytes_xor(a: bytes, b: bytes) -> bytes:  # pylint: disable=invalid-name
    """XOR two bytes values."""
    return (int.from_bytes(a, "big") ^ int.from_bytes(b, "big")).to_bytes(len(a), "big")


def byte_pad(data: bytes, block_size: int) -> bytes:
    """Pad data with 0x00 until its length is a multiple of block_size."""
    if remainder := len(data) % block_size:
        return data + bytes(block_size - remainder)
    return data


def sha256_hash(data: bytes) -> bytes:
    """Calculate SHA256 hash of data."""
    digest = hashes.Hash(hashes.SHA256())
    digest.update(data)
    return digest.finalize()


def calculate_message_authentication_code_cbc(
    key: bytes,
    additional_data: bytes,
    payload: bytes = b"",
    block_0: bytes = bytes(16),
    # counter_0: bytes = bytes(16),
) -> bytes:
    """Calculate the message authentication code (MAC) for a message with AES-CBC."""
    blocks = (
        block_0 + len(additional_data).to_bytes(2, "big") + additional_data + payload
    )
    y_cipher = Cipher(algorithms.AES(key), modes.CBC(bytes(16)))
    y_encryptor = y_cipher.encryptor()
    y_blocks = (
        y_encryptor.update(byte_pad(blocks, block_size=16)) + y_encryptor.finalize()
    )
    # only calculate, no ctr encryption
    return y_blocks[-16:]
    # s_cipher = Cipher(algorithms.AES(key), modes.CTR(counter_0))
    # s_encryptor = s_cipher.encryptor()
    # return s_encryptor.update(y_blocks[-16:]) + s_encryptor.finalize()


def encrypt_data_ctr(
    key: bytes,
    mac_cbc: bytes,
    payload: bytes = b"",
    counter_0: bytes = bytes(16),
) -> bytes:
    """
    Encrypt data with AES-CTR.

    Payload is optional; expected plain KNX/IP frame bytes.
    MAC shall be encrypted with counter 0, KNXnet/IP frame with incremented counters.
    Encrypted MAC is appended to the end of encrypted payload data (if there is any).
    """
    s_cipher = Cipher(algorithms.AES(key), modes.CTR(counter_0))
    s_encryptor = s_cipher.encryptor()
    mac = s_encryptor.update(mac_cbc)
    data = s_encryptor.update(payload) + s_encryptor.finalize()
    return data + mac


def decrypt_ctr(
    session_key: bytes,
    payload: bytes,
    counter_0: bytes = bytes(16),
) -> bytes:
    """
    Decrypt data from SecureWrapper.

    MAC is expected to be the last 16 octets of the payload. This will be sliced and
    decoded first with counter 0.
    Returns a tuple of (KNX/IP frame bytes, MAC TR for verification).
    """
    cipher = Cipher(algorithms.AES(session_key), modes.CTR(counter_0))
    decryptor = cipher.decryptor()
    mac_tr = decryptor.update(payload[-16:])  # MAC is encrypted with counter 0
    decrypted_data = decryptor.update(payload[:-16]) + decryptor.finalize()

    return (decrypted_data, mac_tr)


def calculate_wrapper(
    session_key: bytes,
    encapsulated_frame: bytes,
    secure_session_id: bytes = bytes.fromhex("00 01"),
    sequence_number: bytes = bytes.fromhex("00 00 00 00 00 00"),
    serial_number: bytes = bytes.fromhex("00 fa 12 34 56 78"),
    message_tag: bytes = bytes.fromhex("af fe"),
) -> bytes:
    """Calculate the payload and mac for a secure wrapper."""
    print("# SecureWrapper")

    total_length = (
        6  # KNX/IP Header
        + len(secure_session_id)
        + len(sequence_number)
        + len(serial_number)
        + len(message_tag)
        + len(encapsulated_frame)
        + 16  # MAC
    )
    wrapper_header = bytes.fromhex("06 10 09 50") + total_length.to_bytes(2, "big")

    a_data = wrapper_header + secure_session_id
    p_data = encapsulated_frame
    q_payload_length = len(p_data).to_bytes(2, "big")

    b_0_secure_wrapper = (
        sequence_number + serial_number + message_tag + q_payload_length
    )
    ctr_0_secure_wrapper = (
        sequence_number + serial_number + message_tag + bytes.fromhex("ff") + bytes(1)
    )  # last octet is the counter to increment by 1 each step

    mac_cbc = calculate_message_authentication_code_cbc(
        session_key,
        additional_data=a_data,
        payload=p_data,
        block_0=b_0_secure_wrapper,
    )
    encrypted_data = encrypt_data_ctr(
        session_key,
        mac_cbc=mac_cbc,
        payload=p_data,
        counter_0=ctr_0_secure_wrapper,
    )

    # encrypted data
    # ctr_1_secure_wrapper = (int.from_bytes(ctr_0_secure_wrapper, "big") + 1).to_bytes(
    #     16, "big"
    # )
    # cipher = Cipher(algorithms.AES(session_key), modes.CTR(ctr_1_secure_wrapper))
    # encryptor = cipher.encryptor()
    # enc_frame = encryptor.update(p_data) + encryptor.finalize()

    print(f"encrypted_data: {encrypted_data[16:].hex()}")

    dec_frame, mac_tr = decrypt_ctr(
        session_key,
        payload=encrypted_data,
        counter_0=ctr_0_secure_wrapper,
    )
    assert dec_frame == p_data
    assert mac_tr == mac_cbc  # verification of MAC

    return encrypted_data


def main() -> None:
    """Recalculate KNX specification example frames."""
    ################
    # SessionRequest
    ################
    print("# SessionRequest")
    client_private_key = X25519PrivateKey.from_private_bytes(
        bytes.fromhex(
            "b8 fa bd 62 66 5d 8b 9e 8a 9d 8b 1f 4b ca 42 c8 c2 78 9a 61 10 f5 0e 9d d7 85 b3 ed e8 83 f3 78"
        )
    )
    client_public_key_raw = client_private_key.public_key().public_bytes(
        serialization.Encoding.Raw, serialization.PublicFormat.Raw
    )  # append to SessionRequest (15-46)
    print(f"Public key: {client_public_key_raw.hex()}")

    #################
    # SessionResponse
    #################
    print("# SessionResponse")
    peer_public_key = X25519PublicKey.from_public_bytes(
        bytes.fromhex(
            "bd f0 99 90 99 23 14 3e f0 a5 de 0b 3b e3 68 7b c5 bd 3c f5 f9 e6 f9 01 69 9c d8 70 ec 1f f8 24"
        )
    )
    pub_keys_xor = bytes_xor(
        client_public_key_raw,
        peer_public_key.public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw,
        ),
    )

    peer_device_authentication_password = "trustme"
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=16,
        salt=b"device-authentication-code.1.secure.ip.knx.org",
        iterations=65536,
    )
    # TODO: is encoding "latin-1" correct? (also used for device names in DIBs)
    peer_device_authentication_code = kdf.derive(
        peer_device_authentication_password.encode("latin-1")
    )
    assert peer_device_authentication_code == bytes.fromhex(
        "e1 58 e4 01 20 47 bd 6c c4 1a af bc 5c 04 c1 fc"
    )

    _a_data = bytes.fromhex(
        "06 10 09 52 00 38 00 01 b7 52 be 24 64 59 26 0f 6b 0c 48 01 fb d5 a6 75 99 f8 3b 40 57 b3 ef 1e 79 e4 69 ac 17 23 4e 15"
    )
    ctr_0_session_response = (
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x00"
    )
    message_authentication_code_cbc = calculate_message_authentication_code_cbc(
        peer_device_authentication_code,
        additional_data=_a_data[:8]
        + pub_keys_xor,  # knx_ip_header + secure_session_id + bytes_xor(client_pub_key, server_pub_key)
    )
    message_authentication_code = encrypt_data_ctr(
        peer_device_authentication_code,
        mac_cbc=message_authentication_code_cbc,
        counter_0=ctr_0_session_response,
    )
    assert message_authentication_code == bytes.fromhex(
        "a9 22 50 5a aa 43 61 63 57 0b d5 49 4c 2d f2 a3"
    )

    ecdh_shared_secret = client_private_key.exchange(peer_public_key)
    print(f"ECDH shared secret: {ecdh_shared_secret.hex()}")

    session_key = sha256_hash(ecdh_shared_secret)[:16]
    print(f"Session key: {session_key.hex()}")

    _, mac_tr = decrypt_ctr(
        peer_device_authentication_code,
        payload=message_authentication_code,
        counter_0=ctr_0_session_response,
    )
    assert mac_tr == message_authentication_code_cbc  # verification of MAC

    #####################
    # SessionAuthenticate
    #####################
    #   shall be wrapped in SecureWrapper
    print("# SessionAuthenticate")

    password_string = "secret"
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=16,
        salt=b"user-password.1.secure.ip.knx.org",
        iterations=65536,
    )
    password_hash = kdf.derive(password_string.encode("latin-1"))
    print(f"Password hash: {password_hash.hex(' ')}")
    assert password_hash == bytes.fromhex(
        "03 fc ed b6 66 60 25 1e c8 1a 1a 71 69 01 69 6a"
    )

    authenticate_wrapper = bytes.fromhex(
        "06 10 09 50 00 3e 00 01 00 00 00 00 00 00 00 fa 12 34 56 78 af fe"
        "79 15 a4 f3 6e 6e 42 08"
        "d2 8b 4a 20 7d 8f 35 c0"
        "d1 38 c2 6a 7b 5e 71 69"
        "52 db a8 e7 e4 bd 80 bd"
        "7d 86 8a 3a e7 87 49 de"
    )
    session_authenticate = bytes.fromhex(
        "06 10 09 53 00 18 00 01"
        + "1f 1d 59 ea 9f 12 a1 52 e5 d9 72 7f 08 46 2c de"  # MAC
    )
    mac_cbc_authenticate = calculate_message_authentication_code_cbc(
        password_hash,
        additional_data=session_authenticate[:8] + pub_keys_xor,
        block_0=bytes(16),
    )
    ctr_0_session_authenticate = (
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\x00"
    )
    assert (
        encrypt_data_ctr(
            password_hash,
            mac_cbc=mac_cbc_authenticate,
            counter_0=ctr_0_session_authenticate,
        )
        == session_authenticate[8:]
    )
    assert (
        calculate_wrapper(
            session_key,
            encapsulated_frame=session_authenticate,
            secure_session_id=authenticate_wrapper[6:8],
            sequence_number=authenticate_wrapper[8:14],
            serial_number=authenticate_wrapper[14:20],
            message_tag=authenticate_wrapper[20:22],
        )
        == authenticate_wrapper[22:]
    )
    # verify MAC
    _, mac_tr = decrypt_ctr(
        password_hash,
        payload=session_authenticate[8:],
        counter_0=ctr_0_session_authenticate,
    )
    assert mac_tr == mac_cbc_authenticate

    ###############
    # SessionStatus
    ###############
    #   shall be wrapped in SecureWrapper
    print("# SessionStatus")

    status_wrapper = bytes.fromhex(
        "06 10 09 50 00 2e 00 01 00 00 00 00 00 00 00 fa aa aa aa aa af fe"
        "26 15 6d b5 c7 49 88 8f"
        "a3 73 c3 e0 b4 bd e4 49"
        "7c 39 5e 4b 1c 2f 46 a1"
    )
    session_status = bytes.fromhex("06 10 09 54 00 08 00 00")
    assert (
        calculate_wrapper(
            session_key,
            encapsulated_frame=session_status,
            secure_session_id=status_wrapper[6:8],
            sequence_number=status_wrapper[8:14],
            serial_number=status_wrapper[14:20],
            message_tag=status_wrapper[20:22],
        )
        == status_wrapper[22:]
    )


if __name__ == "__main__":
    main()



"""Unit test for KNX/IP DescriptionRequest objects."""

from xknx.knxip import HPAI, DescriptionRequest, KNXIPFrame


class TestKNXIPDescriptionRequest:
    """Test class for KNX/IP DescriptionRequest objects."""

    def test_description_request(self) -> None:
        """Test parsing and streaming DescriptionRequest KNX/IP packet."""
        raw = bytes.fromhex("06 10 02 03 00 0E 08 01 7F 00 00 02 0E 57")
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, DescriptionRequest)
        assert knxipframe.body.control_endpoint == HPAI(ip_addr="127.0.0.2", port=3671)

        knxipframe2 = KNXIPFrame.init_from_body(
            DescriptionRequest(control_endpoint=HPAI(ip_addr="127.0.0.2", port=3671))
        )

        assert knxipframe2.to_knx() == raw



"""Unit test for KNX 2 byte signed objects."""

import struct
from unittest.mock import patch

import pytest

from xknx.dpt import DPT2ByteSigned, DPTArray
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestDPT2ByteSigned:
    """Test class for KNX 2 byte signed objects."""

    def test_signed_settings(self) -> None:
        """Test members of DPT2ByteSigned."""
        assert DPT2ByteSigned.value_min == -32768
        assert DPT2ByteSigned.value_max == 32767

    def test_signed_assert_min_exceeded(self) -> None:
        """Test initialization of DPT2ByteSigned with wrong value (Underflow)."""
        with pytest.raises(ConversionError):
            DPT2ByteSigned.to_knx(-32769)

    def test_signed_to_knx_exceed_limits(self) -> None:
        """Test initialization of DPT2ByteSigned with wrong value (Overflow)."""
        with pytest.raises(ConversionError):
            DPT2ByteSigned.to_knx(32768)

    def test_signed_value_max_value(self) -> None:
        """Test DPT2ByteSigned parsing and streaming."""
        assert DPT2ByteSigned.to_knx(32767) == DPTArray((0x7F, 0xFF))
        assert DPT2ByteSigned.from_knx(DPTArray((0x7F, 0xFF))) == 32767

    def test_signed_value_min_value(self) -> None:
        """Test DPT2ByteSigned parsing and streaming with null values."""
        assert DPT2ByteSigned.to_knx(-20480) == DPTArray((0xB0, 0x00))
        assert DPT2ByteSigned.from_knx(DPTArray((0xB0, 0x00))) == -20480

    def test_signed_value_0123(self) -> None:
        """Test DPT2ByteSigned parsing and streaming."""
        assert DPT2ByteSigned.to_knx(291) == DPTArray((0x01, 0x23))
        assert DPT2ByteSigned.from_knx(DPTArray((0x01, 0x23))) == 291

    def test_signed_wrong_value_from_knx(self) -> None:
        """Test DPT2ByteSigned parsing with wrong value."""
        with pytest.raises(CouldNotParseTelegram):
            DPT2ByteSigned.from_knx(DPTArray((0xFF, 0x4E, 0x12)))

    def test_from_knx_unpack_error(self) -> None:
        """Test DPT2ByteSigned parsing with unpack error."""
        with patch("struct.unpack") as unpack_mock:
            unpack_mock.side_effect = struct.error()
            with pytest.raises(ConversionError):
                DPT2ByteSigned.from_knx(DPTArray((0x01, 0x23)))

    def test_to_knx_pack_error(self) -> None:
        """Test serializing DPT2ByteSigned with pack error."""
        with patch("struct.pack") as pack_mock:
            pack_mock.side_effect = struct.error()
            with pytest.raises(ConversionError):
                DPT2ByteSigned.to_knx(1234)



"""
Module for managing an RGBW remote value.

DPT 251.600.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from xknx.dpt import DPTArray, DPTBinary, DPTColorRGBW, RGBWColor

from .remote_value import GroupAddressesType, RemoteValue, RVCallbackType

if TYPE_CHECKING:
    from xknx.xknx import XKNX


class RemoteValueColorRGBW(RemoteValue[RGBWColor]):
    """Abstraction for remote value of KNX DPT 251.600 (DPT_Color_RGBW)."""

    def __init__(
        self,
        xknx: XKNX,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        sync_state: bool | int | float | str = True,
        device_name: str | None = None,
        feature_name: str = "Color RGBW",
        after_update_cb: RVCallbackType[RGBWColor] | None = None,
    ) -> None:
        """Initialize remote value of KNX DPT 251.600 (DPT_Color_RGBW)."""
        super().__init__(
            xknx,
            group_address,
            group_address_state,
            sync_state=sync_state,
            device_name=device_name,
            feature_name=feature_name,
            after_update_cb=after_update_cb,
        )
        self._valid_value = RGBWColor()

    def to_knx(self, value: RGBWColor) -> DPTArray | DPTBinary:
        """Convert value to payload."""
        return DPTColorRGBW.to_knx(value)

    def from_knx(self, payload: DPTArray | DPTBinary) -> RGBWColor:
        """
        Convert current payload to value.

        If one element is invalid, use the last received value.
        """
        new_value = DPTColorRGBW.from_knx(payload)
        self._valid_value = self._valid_value | new_value
        return self._valid_value



"""Module for handling CEMI Frames."""

# ruff: noqa: F401
from .cemi_frame import (
    CEMIFrame,
    CEMILData,
    CEMIMPropInfo,
    CEMIMPropReadRequest,
    CEMIMPropReadResponse,
    CEMIMPropWriteRequest,
    CEMIMPropWriteResponse,
)
from .cemi_handler import CEMIHandler
from .const import CEMIErrorCode, CEMIFlags, CEMIMessageCode



"""Unit test for RemoteValueSceneNumber objects."""

import pytest

from xknx import XKNX
from xknx.dpt import DPTArray, DPTBinary
from xknx.exceptions import ConversionError
from xknx.remote_value import RemoteValueTemp
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueWrite


class TestRemoteValueTemp:
    """Test class for RemoteValueTemp objects."""

    def test_to_knx(self) -> None:
        """Test to_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueTemp(xknx)
        assert remote_value.to_knx(11) == DPTArray((0x04, 0x4C))

    def test_from_knx(self) -> None:
        """Test from_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueTemp(xknx)
        assert remote_value.from_knx(DPTArray((0x04, 0x4C))) == 11

    def test_to_knx_error(self) -> None:
        """Test to_knx function with wrong parametern."""
        xknx = XKNX()
        remote_value = RemoteValueTemp(xknx)
        with pytest.raises(ConversionError):
            remote_value.to_knx(-300)
        with pytest.raises(ConversionError):
            remote_value.to_knx("abc")

    def test_process(self) -> None:
        """Test process telegram."""
        xknx = XKNX()
        remote_value = RemoteValueTemp(xknx, group_address=GroupAddress("1/2/3"))
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x04, 0x4C))),
        )
        remote_value.process(telegram)
        assert remote_value.value == 11

    def test_to_process_error(self) -> None:
        """Test process erroneous telegram."""
        xknx = XKNX()
        remote_value = RemoteValueTemp(xknx, group_address=GroupAddress("1/2/3"))

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        assert remote_value.process(telegram) is False

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64,))),
        )
        assert remote_value.process(telegram) is False

        assert remote_value.value is None



"""Abstraction to send DescriptionRequest and wait for DescriptionResponse."""

from __future__ import annotations

from abc import ABC, abstractmethod
import asyncio
import logging
from typing import TYPE_CHECKING, Final

from xknx.exceptions import CommunicationError, XKNXException
from xknx.io import util
from xknx.io.gateway_scanner import GatewayDescriptor
from xknx.knxip import (
    HPAI,
    SRP,
    DescriptionRequest,
    DescriptionResponse,
    DIBTypeCode,
    KNXIPFrame,
    SearchRequestExtended,
    SearchResponseExtended,
)
from xknx.util import asyncio_timeout

from .const import DEFAULT_MCAST_PORT
from .transport import UDPTransport

if TYPE_CHECKING:
    from xknx.io.transport import KNXIPTransport

logger = logging.getLogger("xknx.log")

DESCRIPTION_TIMEOUT: Final = 2


async def request_description(
    gateway_ip: str,
    gateway_port: int = DEFAULT_MCAST_PORT,
    local_ip: str | None = None,
    local_port: int = 0,
    route_back: bool = False,
) -> GatewayDescriptor:
    """Set up a UDP transport to request a description from a KNXnet/IP device."""
    local_ip = local_ip or util.find_local_ip(gateway_ip)
    if local_ip is None:
        # Fall back to default interface and use route back
        local_ip = await util.get_default_local_ip(gateway_ip)
        if local_ip is None:
            raise CommunicationError(
                f"No network interface found to request gateway info from {gateway_ip}:{gateway_port}"
            )
        route_back = True
    try:
        local_ip = await util.validate_ip(local_ip, address_name="Local IP")
        gateway_ip = await util.validate_ip(gateway_ip, address_name="Gateway IP")
    except XKNXException as err:
        raise CommunicationError("Invalid address") from err

    transport = UDPTransport(
        local_addr=(local_ip, local_port),
        remote_addr=(gateway_ip, gateway_port),
        multicast=False,
    )
    try:
        await transport.connect()
    except OSError as err:
        raise CommunicationError(
            "Could not setup socket to request gateway info"
        ) from err
    else:
        local_hpai: HPAI
        if route_back:
            local_hpai = HPAI()
        else:
            local_addr = transport.getsockname()
            local_hpai = HPAI(*local_addr)

        description_query = DescriptionQuery(
            transport=transport,
            local_hpai=local_hpai,
        )
        await description_query.start()
        gateway = description_query.gateway_descriptor
        if gateway is None:
            raise CommunicationError(
                f"Could not fetch gateway info from {gateway_ip}:{gateway_port}"
            )
        if gateway.core_version >= 2:
            search_extended_query = SearchExtendedQuery(
                transport=transport,
                local_hpai=local_hpai,
            )
            await search_extended_query.start()
            gateway = search_extended_query.gateway_descriptor
            if gateway is None:
                raise CommunicationError(
                    f"Could not fetch extended gateway info from {gateway_ip}:{gateway_port}"
                )
        return gateway
    finally:
        transport.stop()


class _SelfDescriptionQuery(ABC):
    """Base class for handling descriptions request-response cycles."""

    expected_response_class: type[DescriptionResponse] | type[SearchResponseExtended]

    def __init__(
        self,
        transport: KNXIPTransport,
        local_hpai: HPAI,
    ) -> None:
        """Initialize Description class."""
        self.transport = transport
        self.local_hpai = local_hpai

        self.gateway_descriptor: GatewayDescriptor | None = None
        self.response_received_event = asyncio.Event()

    @abstractmethod
    def create_knxipframe(self) -> KNXIPFrame:
        """Create KNX/IP Frame object to be sent to device."""

    async def start(self) -> None:
        """Start. Send request and wait for an answer."""
        callb = self.transport.register_callback(
            self.response_rec_callback, [self.expected_response_class.SERVICE_TYPE]
        )
        frame = self.create_knxipframe()
        try:
            self.transport.send(frame)
            async with asyncio_timeout(DESCRIPTION_TIMEOUT):
                await self.response_received_event.wait()
        except asyncio.TimeoutError:
            logger.debug(
                "Error: KNX bus did not respond in time (%s secs) to request of type '%s'",
                DESCRIPTION_TIMEOUT,
                self.__class__.__name__,
            )
        except CommunicationError as err:
            logger.warning("Sending %s failed: %s", frame.body.__class__.__name__, err)
        finally:
            # cleanup to not leave callbacks (for asyncio.CancelledError)
            self.transport.unregister_callback(callb)

    def response_rec_callback(
        self, knxipframe: KNXIPFrame, source: HPAI, _: KNXIPTransport
    ) -> None:
        """Verify and handle knxipframe. Callback from internal transport."""
        if not isinstance(knxipframe.body, self.expected_response_class):
            logger.warning(
                "Wrong knxipframe for %s: %s", self.__class__.__name__, knxipframe
            )
            return
        self.response_received_event.set()
        # Set gateway descriptor attribute
        gateway = GatewayDescriptor(
            ip_addr=self.transport.remote_addr[0],
            port=self.transport.remote_addr[1],
            local_ip=self.transport.getsockname()[0],
        )
        gateway.parse_dibs(knxipframe.body.dibs)  # type: ignore[attr-defined]
        self.gateway_descriptor = gateway


class DescriptionQuery(_SelfDescriptionQuery):
    """Class to send a DescriptionRequest and wait for DescriptionResponse."""

    expected_response_class = DescriptionResponse

    def create_knxipframe(self) -> KNXIPFrame:
        """Create KNX/IP Frame object to be sent to device."""
        description_request = DescriptionRequest(control_endpoint=self.local_hpai)
        return KNXIPFrame.init_from_body(description_request)


class SearchExtendedQuery(_SelfDescriptionQuery):
    """
    Class to send a SearchRequestExtended and wait for SearchResponseExtended to a single device.

    Does only work with UDP transports.
    """

    expected_response_class = SearchResponseExtended

    def create_knxipframe(self) -> KNXIPFrame:
        """Create KNX/IP Frame object to be sent to device."""
        search_extended_request = SearchRequestExtended(
            discovery_endpoint=self.local_hpai,
            srps=[
                SRP.request_device_description(
                    [
                        DIBTypeCode.DEVICE_INFO,
                        DIBTypeCode.SUPP_SVC_FAMILIES,
                        DIBTypeCode.SECURED_SERVICE_FAMILIES,
                        DIBTypeCode.TUNNELING_INFO,
                    ]
                )
            ],
        )
        return KNXIPFrame.init_from_body(search_extended_request)



"""Encryption and Decryption functions for KNX Secure."""

from __future__ import annotations

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from .util import Buffer, byte_pad


def calculate_message_authentication_code_cbc(
    key: bytes,
    additional_data: bytes,
    payload: Buffer = b"",
    block_0: bytes = bytes(16),
) -> bytes:
    """Calculate the message authentication code (MAC) for a message with AES-CBC."""
    blocks = (
        block_0 + len(additional_data).to_bytes(2, "big") + additional_data + payload
    )
    y_cipher = Cipher(algorithms.AES(key), modes.CBC(bytes(16)))
    y_encryptor = y_cipher.encryptor()
    y_blocks = (
        y_encryptor.update(byte_pad(blocks, block_size=16)) + y_encryptor.finalize()
    )
    # only calculate, no ctr encryption
    return y_blocks[-16:]


def decrypt_ctr(
    key: bytes,
    counter_0: bytes,
    mac: bytes,
    payload: Buffer = b"",
) -> tuple[bytes, bytes]:
    """
    Decrypt data from SecureWrapper.

    MAC will be decoded first with counter 0.
    Returns a tuple of (KNX/IP frame bytes, MAC TR for verification).
    """
    cipher = Cipher(algorithms.AES(key), modes.CTR(counter_0))
    decryptor = cipher.decryptor()
    mac_tr = decryptor.update(mac)  # MAC is encrypted with counter 0
    decrypted_data = decryptor.update(payload) + decryptor.finalize()

    return (decrypted_data, mac_tr)


def encrypt_data_ctr(
    key: bytes,
    counter_0: bytes,
    mac_cbc: bytes,
    payload: Buffer = b"",
) -> tuple[bytes, bytes]:
    """
    Encrypt data with AES-CTR.

    Payload is expected a full Plain KNX/IP frame with header.
    MAC shall be encrypted with counter 0, KNXnet/IP frame with incremented counters.
    Returns a tuple of encrypted data (if there is any) and encrypted MAC.
    """
    s_cipher = Cipher(algorithms.AES(key), modes.CTR(counter_0))
    s_encryptor = s_cipher.encryptor()
    mac = s_encryptor.update(mac_cbc)
    encrypted_data = s_encryptor.update(payload) + s_encryptor.finalize()
    return (encrypted_data, mac)


def derive_device_authentication_password(device_authentication_password: str) -> bytes:
    """Derive device authentication password."""
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=16,
        salt=b"device-authentication-code.1.secure.ip.knx.org",
        iterations=65536,
    )
    return kdf.derive(device_authentication_password.encode("latin-1"))


def derive_user_password(password_string: str) -> bytes:
    """Derive user password."""
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=16,
        salt=b"user-password.1.secure.ip.knx.org",
        iterations=65536,
    )
    return kdf.derive(password_string.encode("latin-1"))


def generate_ecdh_key_pair() -> tuple[X25519PrivateKey, bytes]:
    """
    Generate an ECDH key pair.

    Return the private key and the raw bytes of the public key.
    """
    private_key = X25519PrivateKey.generate()
    public_key_raw = private_key.public_key().public_bytes(
        serialization.Encoding.Raw, serialization.PublicFormat.Raw
    )
    return private_key, public_key_raw



"""Unit test for TPCI objects."""

import pytest

from xknx.exceptions import ConversionError
from xknx.telegram.tpci import (
    TPCI,
    TAck,
    TConnect,
    TDataBroadcast,
    TDataConnected,
    TDataGroup,
    TDataIndividual,
    TDataTagGroup,
    TDisconnect,
    TNak,
)


@pytest.mark.parametrize(
    ("tpci_int", "dst_is_group_address", "dst_is_zero", "tpci_expected"),
    [
        (0b00000000, True, False, TDataGroup()),
        (0b00000000, True, True, TDataBroadcast()),
        (0b00000100, True, False, TDataTagGroup()),
        (0b00000000, False, False, TDataIndividual()),
        (0b00000000, False, True, TDataIndividual()),
        (0b01011100, False, False, TDataConnected(sequence_number=0b0111)),
        (0b10000000, False, False, TConnect()),
        (0b10000000, False, True, TConnect()),
        (0b10000001, False, False, TDisconnect()),
        (0b11101010, False, False, TAck(sequence_number=10)),
        (0b11010011, False, False, TNak(sequence_number=4)),
    ],
)
def test_tpci_resolve_encode(
    tpci_int: int, dst_is_group_address: bool, dst_is_zero: bool, tpci_expected: TPCI
) -> None:
    """Test resolving and encoding TPCI classes."""
    assert (
        TPCI.resolve(
            raw_tpci=tpci_int,
            dst_is_group_address=dst_is_group_address,
            dst_is_zero=dst_is_zero,
        )
        == tpci_expected
    )
    assert tpci_expected.to_knx() == tpci_int


@pytest.mark.parametrize("dst_is_zero", [True, False])
@pytest.mark.parametrize(
    ("tpci_int", "dst_is_group_address"),
    [
        # sequence_number for non-numbered
        (0b00100000, True),
        (0b00001000, False),
        (0b10000100, True),
        (0b10000100, False),
        # numbered for group addressed
        (0b01000000, True),
    ],
)
def test_invalid_tpci(
    tpci_int: int, dst_is_group_address: bool, dst_is_zero: bool
) -> None:
    """Test resolving TPCI classes."""
    with pytest.raises(ConversionError):
        TPCI.resolve(
            raw_tpci=tpci_int,
            dst_is_group_address=dst_is_group_address,
            dst_is_zero=dst_is_zero,
        )


def test_equality() -> None:
    """Test equality."""
    assert TDataGroup() == TDataGroup()
    assert TDataGroup() != TDataIndividual()



"""
Module for Serialization and Deserialization of a KNX Connetionstate Request information.

Connectionstate requests are used to determine if a tunnel connection is still active and valid.
"""

from __future__ import annotations

from xknx.exceptions import CouldNotParseKNXIP

from .body import KNXIPBody
from .hpai import HPAI
from .knxip_enum import KNXIPServiceType


class ConnectionStateRequest(KNXIPBody):
    """Representation of a KNX Connection State Request."""

    SERVICE_TYPE = KNXIPServiceType.CONNECTIONSTATE_REQUEST

    def __init__(
        self,
        communication_channel_id: int = 1,
        control_endpoint: HPAI | None = None,
    ) -> None:
        """Initialize ConnectionStateRequest object."""
        self.communication_channel_id = communication_channel_id
        self.control_endpoint = control_endpoint or HPAI()

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return 2 + HPAI.LENGTH

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""

        def info_from_knx(info: bytes) -> int:
            """Parse info bytes."""
            if len(info) < 2:
                raise CouldNotParseKNXIP("Info has wrong length")
            self.communication_channel_id = info[0]
            # info[1] is reserved
            return 2

        pos = info_from_knx(raw)
        pos += self.control_endpoint.from_knx(raw[pos:])
        return pos

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return (
            bytes((self.communication_channel_id, 0x00))  # 2nd byte is reserved
            + self.control_endpoint.to_knx()
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            "<ConnectionStateRequest "
            f'communication_channel_id="{self.communication_channel_id}", '
            f'control_endpoint="{self.control_endpoint}" />'
        )



"""Unit test for KNX string object."""

import pytest

from xknx.dpt import DPTArray, DPTLatin1, DPTString
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestDPTString:
    """Test class for KNX ASCII string object."""

    @pytest.mark.parametrize(
        "string,raw",
        [
            (
                "KNX is OK",
                (75, 78, 88, 32, 105, 115, 32, 79, 75, 0, 0, 0, 0, 0),
            ),
            (
                "",
                (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
            ),
            (
                "AbCdEfGhIjKlMn",
                (65, 98, 67, 100, 69, 102, 71, 104, 73, 106, 75, 108, 77, 110),
            ),
            (
                ".,:;-_!?$@&#%/",
                (46, 44, 58, 59, 45, 95, 33, 63, 36, 64, 38, 35, 37, 47),
            ),
        ],
    )
    @pytest.mark.parametrize("test_dpt", [DPTString, DPTLatin1])
    def test_values(
        self, string: str, raw: tuple[int, ...], test_dpt: type[DPTString]
    ) -> None:
        """Test parsing and streaming strings."""
        assert test_dpt.to_knx(string) == DPTArray(raw)
        assert test_dpt.from_knx(DPTArray(raw)) == string

    @pytest.mark.parametrize(
        "string,knx_string,raw",
        [
            (
                "Matouš",
                "Matou?",
                (77, 97, 116, 111, 117, 63, 0, 0, 0, 0, 0, 0, 0, 0),
            ),
            (
                "Gänsefüßchen",
                "G?nsef??chen",
                (71, 63, 110, 115, 101, 102, 63, 63, 99, 104, 101, 110, 0, 0),
            ),
        ],
    )
    def test_to_knx_ascii_invalid_chars(
        self, string: str, knx_string: str, raw: tuple[int, ...]
    ) -> None:
        """Test streaming ASCII string with invalid chars."""
        assert DPTString.to_knx(string) == DPTArray(raw)
        assert DPTString.from_knx(DPTArray(raw)) == knx_string

    @pytest.mark.parametrize(
        "string,raw",
        [
            (
                "Gänsefüßchen",
                (71, 228, 110, 115, 101, 102, 252, 223, 99, 104, 101, 110, 0, 0),
            ),
            (
                "àáâãåæçèéêëìíî",
                (224, 225, 226, 227, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238),
            ),
        ],
    )
    def test_to_knx_latin_1(self, string: str, raw: tuple[int, ...]) -> None:
        """Test streaming Latin-1 strings."""
        assert DPTLatin1.to_knx(string) == DPTArray(raw)
        assert DPTLatin1.from_knx(DPTArray(raw)) == string

    def test_to_knx_too_long(self) -> None:
        """Test serializing DPTString to KNX with wrong value (to long)."""
        with pytest.raises(ConversionError):
            DPTString.to_knx("AAAAABBBBBCCCCx")

    @pytest.mark.parametrize(
        "raw",
        [
            ((0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),),
            ((0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),),
        ],
    )
    def test_from_knx_wrong_parameter_length(self, raw: tuple[int, ...]) -> None:
        """Test parsing of KNX string with wrong elements length."""
        with pytest.raises(CouldNotParseTelegram):
            DPTString.from_knx(DPTArray(raw))

    def test_no_unit_of_measurement(self) -> None:
        """Test for no unit set for DPT 16."""
        assert DPTString.unit is None
